/*
** Restricted Rights Legend
**
** Use, duplication or disclosure by the government is subject to restrictions
** as set forth in subparagraph (c)(1)(ii) of the Rights in Technical Data and
** Computer Software clause at DFARS 252.227-7013.
**
** Copyright (c) 1997 Lockheed Martin C2 Integration Systems
**
**
** File:         KqmlAdmin.java
**
** Description:  Contains the KqmlAdmin class that is internal to the KRIL.
**
*/
package kqml;
import java.util.Vector;

/**
 * A class internal to the KRIL.
 *
 * This class is run as a thread and used to perform KQML administrative
 * work, such as sending out timing messages.
 *
 * @author Peter Stevens
 * @version $Revision: 1.6 $
 * @see KqmlAgent
 */
class KqmlAdmin extends Thread
{
    private KqmlAgent agent    = null;
    private MsgQueue  msgQueue = null;
    private long      timeout  = 0;
    private Vector    work     = null;
    private Boolean   adminActive = new Boolean(false);

    private static final long SLEEP_TIME = 50;
    private static final long PENDING_MSGS_TIME_INTERVAL = 100000;

    // Constructor
    KqmlAdmin(KqmlAgent agent)
    {
        super("Admin");
        this.agent = agent;
        work       = new Vector();
        msgQueue   = new MsgQueue(this);
    }

    public void run()
    {
        Log.log(Log.THREAD, "Agent " + agent.getAgentName() +
		" starting KqmlAdmin thread `" + this.getName() + "'");
        agent.addLookupThread(this);    // for sendWithReply operations
        setPriority(getPriority() + 1);

        while (true)
        {
            // Get a message from this thread's unsolicited work queue
            Object obj = getWork();

            if (obj != null)
            {
                setAdminActive(true);
                if (obj instanceof TimingRecord)
                    ((TimingRecord)obj).sendTiming(agent);
                else
                    Log.log(Log.ERROR,
			    "Work queue for KqmlAdmin thread in agent " +
			    agent.getAgentName() +
			    "\n received and ignoring(invalid type) object " +
			    obj);
                setAdminActive(false);
            }
        }
    }

    //
    // These are for queueing "KqmlAdmin" work items
    //
    private Object getWork()
    {
        Object rval = null;

        synchronized (work)
        {
            if (work.isEmpty())
                rval = null;
            else
            {
                rval = work.firstElement();
                work.removeElement(rval);
            }
        }
        if (rval == null)
        {
            this.suspend();
            synchronized (work)
            {
                if (!work.isEmpty())
                {
                    rval = work.firstElement();
                    work.removeElement(rval);
                }
            }
        }
        return (rval);
    }

    // Visible to the KRIL
    protected void putWork(Object obj)
    {
        synchronized (work)
        {
            work.addElement(obj);
        }
        if (getAdminActive() == false)
            this.resume();
    }

    protected void makePending(Message msg)
    {
        //
        // Register the thread as waiting for this message
        //

        //
        // Note this thread is waiting for this message
        //
        agent.pendingMsgPutThread(msg, this);

	// Since this message is now on the pendingMsgs queue, it can be
	// accessed by the checkTimeouts() method of the KqmlAgent object
	// which will check to see if it has timed out.  So, we must reset
	// the timeout value for this thread here.  Since we don't want to
	// set that value for real until getExpectedMsg() is called, set it
	// to something big enough that it will allow for the different 
	// threads to catch up to each other.
	this.timeout = System.currentTimeMillis() + PENDING_MSGS_TIME_INTERVAL;

        Log.log(Log.IO+Log.DEBUG, "KqmlAdmin:makePending msgId `" +
                msg.getKqmlMsgId() + "'\n     for dispatch thread `" +
                this.getName() + "'");
    }

    //
    // The follwing two methods are needed by the KRIL to support
    // non-blocking message reception in this type of KQML Thread
    // (i.e. KqmlAdmin)
    //
    protected Message getExpectedMsg(long timeout)
    {
        //
        // The timeout value is computed here as a time in the future
        // to save having to compute it every time the timeouts are
        // checked in KqmlAgent.checkTimeouts().
        //
        this.timeout = System.currentTimeMillis() + timeout;
	Log.log(Log.DEBUG, "getExpectedMsg: set timeout=" + this.timeout +
		" for thread `" + this.getName() + "'");

        //
        // Block this thread, waiting for the message
        //
        return (msgQueue.getExpectedMsg());
    }

    protected void putExpectedMsg(Message msg) { msgQueue.putExpectedMsg(msg); }

    protected Thread getThread()  { return ((Thread)this); }
    protected long   getTimeout() { return (timeout); }

    private boolean getAdminActive()
    {
        boolean b;
        synchronized (adminActive)
        {
            b = adminActive.booleanValue();
        }
        return (b);
    }

    private void setAdminActive(boolean b)
    {
        synchronized (adminActive)
        {
            adminActive = new Boolean(b);
        }
    }
}
