/*
** Restricted Rights Legend
**
** Use, duplication or disclosure by the government is subject to restrictions
** as set forth in subparagraph (c)(1)(ii) of the Rights in Technical Data and
** Computer Software clause at DFARS 252.227-7013.
**
** Copyright (c) 1997 Lockheed Martin C2 Integration Systems
**
**
** File:         KqmlAgent.java
**
** Description:  Contains the KqmlAgent class with the following "public"
**               methods:
**
**                   Constructors to create an agent instance and support 
**                   threads (MsgReceiver, KqmlDispatch, KqmlAdmin).
**                   Methods that enable or disable accepting
**                   "kqml-admin :content (shutdown)" requests, shutdown an
**                   agent,register an unsolicited message handler, register
**                   an application thread so that Message.sendAndAwaitReply
**                   method can be used (performed automatically for the 
**                   thread that creates the agent instance), get the agent
**                   instance's name, a method to determine if an agent
**                   instance has been shutdown, methods to set and get
**                   the agent's default timeout value, methods to enable,
**                   disable and query message timing collection state.
*/

package kqml;
import java.net.InetAddress;
import java.net.ServerSocket;
import java.io.EOFException;
import java.net.UnknownHostException;
import java.util.Date;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;

/**
 * A class to create an instance of a KqmlAgent.  Upon creation of a
 * KqmlAgent instance, there will be a message receiver thread, a new
 * connection listener thread, a kqml admin thread and at least one
 * message handler thread (the default number is 5 message handler threads)
 * created to invoke an performative's message handler.  The following
 * code fragment will create an agent <i>myClient</i> at the ANS specified
 * in the ANS_HOST system property or at the local host's ANS if no
 * ANS_HOST system property is defined.
 *
 * <code>
 * KqmlAgent agent;
 * try
 * {
 *    agent = new KqmlAgent("myClient");
 * }
 * catch (KqmlAgentInitFailure e)
 * {
 *    System.out.println(e.getMessage());
 * }
 * catch (KqmlAgentRegistrationFailure e)
 * {
 *    System.out.println(e.getMessage());
 * }</code>
 * 
 * @author Peter Stevens
 * @version $Revision: 1.8 $
 * @see Message
 * @see Reply
 */
public class KqmlAgent extends Thread implements KqmlKeywords
{
    //
    // Constants
    //
    private static final int    ANS_LOOKUP_TIMEOUT        = 3000;
    private static final int    DEFAULT_AGENT_THREADS     = 5;
    private static final int    DEFAULT_TIMEOUT           = 30000;
    private static final int    KQML_ANS_PORT             = 5500;
    private static final int    KQML_ACCEPT_QUEUE_DEPTH   = 50;
    private static final int    MILLISECONDS_PER_SEC      = 1000;

    private static final String ASK_ONE                   = "ask-one";
    private static final String DESTINATION_AGENT_UNKNOWN =
                                              "destination agent unknown `";
    private static final String KQML_ADMIN                = "kqml-admin";
    private static final String KQML_PROTOCOL             = "tcp-ip";
    private static final String NAME                      = "(name)";
    private static final String SHUTDOWN                  = "(shutdown)";
    private static final String TIMINGS_OFF               = "(timings off)";
    private static final String TIMINGS_ON                = "(timings on)";
    private static final String LOGGING_ON                = "(logging on)";
    private static final String LOGGING_OFF               = "(logging off)";


    //
    // Private Data
    //
    private KqmlAdmin        admin            = null;
    private String           agentAddr        = null;
    private Hashtable        agentLookup      = new Hashtable();
    private String           agentName        = null;
    private boolean          agentState       = false;
    private Connection       ans              = null;
    private String           ansAgent         = null;
    private Vector           connCache        = new Vector();
    private Hashtable        dispatchTable    = new Hashtable();
    private int              lastCacheSize    = 0;
    private String           listenerAddr     = null;
    private int              listenerPort     = 0;
    private KqmlDispatch     msgThreads[]     = null;
    private int              noThreads        = 0;
    private Hashtable        notRegistered    = new Hashtable();
    private Hashtable        pendingMsgs      = new Hashtable();
    private boolean          permitShutdown   = true;
    private MsgReceiver      receiver         = null;
    private ServerSocket     server           = null;
    private int              serverPort       = 0;
    private Hashtable        threadLookup     = new Hashtable();
    private int              timeout          = getKqmlTimeout();
    private boolean          timingEnabled    = false;
    private Hashtable        timingTable      = new Hashtable();
    private boolean          loggingEnabled   = false;
    private Hashtable        loggingTable     = new Hashtable();
    private UnsolicitedQueue unsolicitedQueue = new UnsolicitedQueue();
    private Vector           userThread       = new Vector();

    //
    // Constructors
    //
    /**
     * Returns a KqmlAgent instance with the agent registered as
     * the agent name specified in agentName parameter and the default
     * number of message handler threads (5), at the default ANS (the
     * local host) or with the ANS specified in the ANS_HOST system
     * property.
     *
     * @param agentName   the agent's name for registration at the ANS
     * @exception KqmlAgentInitFailure          failed to obtain a Server socket
     * @exception KqmlAgentRegistrationFailure  registration failed at the ANS
     */
    public KqmlAgent(String agentName)
                throws KqmlAgentInitFailure, KqmlAgentRegistrationFailure
    {
        this(agentName, DEFAULT_AGENT_THREADS, KqmlAgent.getAnsHost());
    }

    /**
     * Returns a KqmlAgent instance with the agent registered as the
     * agent name specified in the input parameter and with the programmer
     * specified number of message handler threads, at the default 
     * ANS (the local host) or with the ANS specified in the ANS_HOST
     * system property.
     *
     * @param agentName   the agent's name for registration at the ANS
     * @param noThreads   the desired number of message handler threads
     * @exception KqmlAgentInitFailure          failed to obtain a Server socket
     * @exception KqmlAgentRegistrationFailure  less than one Message Handler
     *                                          thread specified or
     *                                          registration failed at the ANS
     */
    public KqmlAgent(String agentName, int noThreads)
                throws KqmlAgentInitFailure, KqmlAgentRegistrationFailure
    {
        this(agentName, noThreads, KqmlAgent.getAnsHost());
    }

    /**
     * Returns a KqmlAgent instance with the agent registered as the
     * agent name specified in agentName parameter and the default number
     * of message handler threads (5), at the programmer specified ANS.
     *
     * @param agentName   the agent's name for registration at the ANS
     * @param ansName     the desired ANS host, assumed to be socket 5500
     *                    on the named host.
     * @exception KqmlAgentInitFailure          failed to obtain a Server socket
     * @exception KqmlAgentRegistrationFailure  less than one Message Handler
     *                                          thread specified or
     *                                          registration failed at the ANS
     */
    public KqmlAgent(String agentName, String ansName)
                throws KqmlAgentInitFailure, KqmlAgentRegistrationFailure
    {
        this(agentName, DEFAULT_AGENT_THREADS, ansName);
    }

    /**
     * Returns a KqmlAgent instance with the agent registered as the
     * agent name specified in agentName parameter and the programmer
     * specified number of message handler threads, at the programmer
     * specified ANS.
     *
     * @param agentName   the agent's name for registration at the ANS
     * @param noThreads   the desired number of message handler threads
     * @param ansName     the desired ANS host, assumed to be socket 5500
     *                    on the named host.
     * @exception KqmlAgentInitFailure          failed to obtain a Server socket
     * @exception KqmlAgentRegistrationFailure  less than one Message Handler
     *                                          thread specified or
     *                                          registration failed at the ANS
     */
    public KqmlAgent(String agentName, int noThreads, String ansHost)
                throws KqmlAgentInitFailure, KqmlAgentRegistrationFailure
    {
        super(agentName);

        if (noThreads < 1)
            throw new KqmlAgentRegistrationFailure(
                    "Must register agent " + agentName +
                    " with at least one thread, noThreads was " + noThreads);

        this.noThreads = noThreads;
        this.agentName = agentName;

        //
        // Establish a "listener" port for accepting connections
        // from outsiders.
        //
        try
        {
            String  contents;
            String  performative;
            Message register;
            Message response;

            //
            // ServerSocket can throw java.io.IOException
            //
            server     = new ServerSocket(0, KQML_ACCEPT_QUEUE_DEPTH);
            serverPort = server.getLocalPort();

            //
            // Record listener address information and the agent address
            // string for the "register" performative
            //
            listenerAddr  = InetAddress.getLocalHost().getHostAddress();
            listenerPort  = server.getLocalPort();
            agentAddr     = listenerAddr + ":" + listenerPort + ":";

            //
            // Establish the connection to the ANS for the "register"
            //
            ansAgent      = "ANS-" + ansHost;
            ans           = new Connection(ansHost, KQML_ANS_PORT,
                                                            ansAgent, this);
            connCache.addElement(ans);  // has to be in the cache for
                                        // for listening

            registerThread();           // register this thread for
                                        // use in sendAndAwaitReply()

            //
            // Create the "register" message, send it to the ANS, get
            // the response back, and verify a successful registration
            //
            register      = new Message(ans, REGISTER, "(" + agentName
                                        + " " + agentAddr + ")", ansAgent);

            response      = register.blockingSendAndAwaitReply();
            contents      = response.getContents();

            if (response.perfIsReply())
            {
                Log.log(Log.READ+Log.DEBUG, "Registered " + agentName +
			" AT " + contents);
            }
            else
            {
                Log.log(Log.ERROR, "Register " + agentName + " FAILED `" +
                                contents + "'");
                throw new KqmlAgentRegistrationFailure(
                    "Register agent response is " +
                    response.getPerformative() + " " + contents);
            }

            this.start();       // Start the listener thread

            //
            // Allocate and start the message dispatching threads
            //
            msgThreads = new KqmlDispatch[noThreads];
            for (int i = 0; i < noThreads; ++i)
            {
                msgThreads[i] = new KqmlDispatch(this, unsolicitedQueue);
                msgThreads[i].start();
            }

            //
            // Create an instance of and start the Msg Receiver
            // since this will "start" the thread that "interprets"
            // incoming messages, do this after listener and handler 
            // threads have been started.
            //
            receiver = new MsgReceiver(this);
            receiver.start();   // Start the message receiver thread

            //
            // Setup and start the "KqmlAdmin" thread
            //
            admin = new KqmlAdmin(this);
            admin.start();      // Start the admin thread

            agentState = true;
        }
        catch (java.io.IOException e)
        {
            throw new KqmlAgentInitFailure(
                "Unable to establish kqml listener for agent " + agentName);
        }
        catch (KqmlIOException e)
        {
            Log.log(Log.ERROR, "Register Agent IO EXCEPTION SENDING REGISTER" +
		    " TO ANS: " + e.getMessage());
            throw new KqmlAgentRegistrationFailure(
                "Register agent response msg exception");
        }
        catch (InvalidKQMLOpenException e)
        {
            Log.log(Log.ERROR, "Register Agent CAN'T OPEN SOCKET TO ANS: " +
		    e.getMessage());
            throw new KqmlAgentRegistrationFailure(
                "Register agent response msg exception");
        }
        catch (InvalidKQMLMsgException e)
        {
            Log.log(Log.ERROR, "Register Agent INVALID MSG: " +
		    e.getMessage());
            throw new KqmlAgentRegistrationFailure(
                "Register agent response msg exception");
        }
        catch (InvalidKQMLTokenException e)
        {
            Log.log(Log.ERROR, "Register Agent INVALID TOKEN: " +
		    e.getMessage());
            throw new KqmlAgentRegistrationFailure(
                "Register agent response msg exception");
        }
        catch (KQMLMsgTimeout e)
        {
            Log.log(Log.ERROR, "Register Agent MESSAGE TIMEOUT: " +
		    e.getMessage());
            throw new KqmlAgentRegistrationFailure(
                "Register agent reply timeout");
        }
    }

    /**
     * Generates a string representing a KqmlAgent that includes the
     * agent name, host and listening port.
     * @return KqmlAgent string which includes that name, host, and listening
     * port
     */
    public String toString()
    {
        return (agentName + ":" + agentAddr);
    }

    /**
     * Sets the permitShutdown state for this agent.  This state variable
     * controls whether a <code>kqml-admin :content (shutdown)</code>
     * KQML message will shutdown the agent.  The default setting at
     * agent creation is true.
     *
     * @param state   the new setting for permitShutdown
     */
    public void setShutdownState(boolean state)
    {
        permitShutdown = state;
    }

    //
    // Cleanup 
    //
    /**
     * Performs a graceful shutdown of this agent by sending an "unregister"
     * message to the ANS where this agent is registered, closing all open
     * sockets and stopping the listener thread, the message receiver
     * thread, the kqml admin thread, and all message handler threads
     * (except if a message handler thread is the current thread, it isn't
     * stopped).  After a shutdown, the agent's active state is set to false.
     *
     * <b>Note: it is strongly recommended that an agent cleanly shutdown
     * because OS generated external signals are not propagated by the
     * Java Virtual Machine (JVM) and thus can't be handled in the Java
     * application.  The JVM will simply terminate without any notice to
     * the Java application.  Thus, if the application is normally terminated
     * by a keyboard "interrupt" (e.g. Control-C on Unix) it will remain
     * registered at the ANS and none of the sockets will be closed.</b>
     *
     * @see #agentActive()
     * @see #setShutdownState()
     */
    public void shutdown()
    {
        Thread myThread = Thread.currentThread();
        int myPriority  = myThread.getPriority();

        //
        // Thread.stop() doesn't seem to work on a higher priority thread
        // (by empirical observation), so set the priority of this thread
        // to the maximum.  It is reset to its original value at the end
        // of this routine.
        //
        myThread.setPriority(Thread.MAX_PRIORITY);

        //
        // Inform the ANS we're going away 
        //
        try
        {
            Message msg = new Message(ans, UNREGISTER, agentName, ansAgent);
            msg.send();
            Thread.sleep(300);      // Wait awhile for the "unregister" to ship
        }
        catch (KqmlException e)
        {
            System.err.println("KqmlAgent shutdown " + e.getMessage());
        }
        catch (Throwable e)
        {
            System.err.println("KqmlAgent shutdown " + e.getMessage());
        }

        //
        // Stop the listener thread and close the listener socket
        //
        try
        {
            if (!super.equals(myThread))
            {
                this.stop();
                Log.log(Log.IO+Log.DEBUG, "shutdown: listener thread stopped");
            }
        }
        catch (SecurityException e)
        {
        }
        try
        {
            server.close();
            Log.log(Log.IO+Log.DEBUG,
                    "shutdown: closed listener socket at port " + serverPort);
        }
        catch (java.io.IOException e)
        {
        }

        //
        // Stop the receiver thread and close the connection cache sockets
        //
        try
        {
            if (!receiver.equals(myThread))
            {
                receiver.stop();
                Log.log(Log.IO+Log.DEBUG, "shutdown: receiver thread stopped");
            }
        }
        catch (SecurityException e)
        {
        }

        Log.log(Log.IO+Log.DEBUG, "shutdown: connection cache has " +
		connCache.size() + " elements");
        for (int i = 0; i < connCache.size(); ++i)
        {
            ((Connection)connCache.elementAt(i)).close();
        }

        //
        // Stop the handler threads
        //
        for (int i = 0; i < noThreads; ++i)
        {
            try
            {
                if (!msgThreads[i].equals(myThread))
                {
                    msgThreads[i].stop();
                    Log.log(Log.IO+Log.DEBUG, "shutdown: handler thread " + i +
                            " stopped");
                }
            }
            catch (SecurityException e)
            {
            }
        }

        //
        // Stop the admin thread
        //
        try
        {
            //
            // set the priority low and resume it, then stop it
            // - stop doesn't seem to work on a suspended thread.
            //
            admin.resume();
            admin.stop();
            Log.log(Log.IO+Log.DEBUG, "shutdown: admin thread stopped");
        }
        catch (SecurityException e)
        {
        }

        myThread.setPriority(myPriority);
        agentState = false;
    }

    //
    // The application has to register any thread that wishes to use
    // "sendAndAwaitReply" so that the KRIL will supply getExpectedMsg
    // and putExpectedMsg methods.
    //
    // Note:   The application thread that creates the KqmlAgent instance
    //         is registered automatically by creation of the agent
    //
    /**
     * Registers an application thread so that it may properly use the Message
     * <code>sendAndAwaitReply()</code> method.
     *
     * <b>Note:  this is not needed for the thread that creates the KqmlAgent
     * instance because it is automatically registered by creation of the 
     * instance.</b>
     *
     * <b>Note:  failure to register an application thread (other than the
     * the thread creating the agent instance) and subsequent use of the
     * </b><code>sendAndAwaitReply()</code><b> method by such a thread is not
     * be detected by the KRIL and will likely result in undefined behavior
     * in such an application thread.</b>
     *
     * @see Message#sendAndAwaitReply()
     * @see Message#sendAndAwaitReply(long)
     */
    public void registerThread()
    {
        userThread.addElement(new KqmlRegistration(this));
    }

    //
    // Application enters a performative name (string) in the
    // agent's performative dispatch table
    //
    // Note: performative names are translated to lower case for case 
    //       insensitive handling
    //
    /**
     * Registers an application's handler for the argument performative.
     * The user's Handler object extends the Handler abstract class.
     *
     * Performative comparison is <b>case insenstive</b> so that
     * registering for the performative <code>ask-all</code> implicitly
     * registers that handler for the <code>ASK-ALL</code> performative.
     *
     * @param perf   the performative name for which the handler 
     *               object is registered.
     * @param h      the handler object whose <code>handler</code>
     *               is invoked when a message with the specified
     *               performative is received.
     * @see Handler
     */
    public synchronized void registerHandler(String perf, Handler h)
    {
        dispatchTable.put(perf.toLowerCase(), h);
    }

    //
    // Handler Dispatching Support
    //
    // Note: performative names are translated to lower case for case 
    //       insensitive handling
    //
    protected void dispatch(Message msg)
    {
        Handler h;
        String  perf  = msg.getPerformative();
        Reply   reply = null;

        //
        // kqml-admin performative operations are handled by the KRIL
        // directly
        //
        if (perf.equalsIgnoreCase(KQML_ADMIN))
        {
            String  content = msg.getContents().toLowerCase();

            if (content.equals(NAME))
            {
                reply = Reply.reply(agentName);
                Log.log(Log.WRITE+Log.DEBUG, "kqml-admin name request from " +
                            msg.getSender());
            }
            else if (content.equals(SHUTDOWN))
            {
                if (permitShutdown)
                {
                    Log.log(Log.WRITE+Log.DEBUG,
			    "kqml-admin shutdown request from " +
			    msg.getSender());
                    shutdown();
                    System.exit(0);
                }
                else
                {
                    reply = Reply.sorry(
                                "kqml-admin shutdown requests are disabled");
                    Log.log(Log.ERROR,
			    "ignored kqml-admin shutdown request from " +
			    msg.getSender());
                }
            }
            else if (content.equals(TIMINGS_OFF))
            {
                disableTiming();
            }
            else if (content.equals(TIMINGS_ON))
            {
                enableTiming();
            }
            else if (content.equals(LOGGING_OFF))
            {
                disableLogging();
            }
            else if (content.equals(LOGGING_ON))
            {
                enableLogging();
            }
            else
            {
                reply = Reply.sorry("Unsupported kqml-admin request `" +
				    content + "'");
                Log.log(Log.ERROR, "Unsupported kqml-admin request `" + 
			content + "'");
            }
        }
        else
        {
            //
            // Fetch the handler class pointer
            //
            synchronized (dispatchTable)
            {
                h = (Handler)dispatchTable.get(perf);
            }

            //
            // If the pointer is null, there isn't a handler for this message's
            // performative and generate an "error" response to the sender.
            // Otherwise, invoke the handler object's "handler" method
            //

            if (h == null)
                reply = Reply.error("Can't handle performative `" + perf + "'");
            else
            {
                //
                // Try and respond gracefully to an exception that isn't
                // caught in the user code
                //
                try
                {
                    reply = h.handler(msg);
                }
                catch (Exception e)
                {
                    reply = Reply.sorry("Application exception: " +
					e.getMessage());
                    Log.logStackTrace(Log.ERROR, e);
                }
            }
        }

        if (reply != null)
        {
            Log.log(Log.WRITE+Log.DEBUG, ">> call   sendResponse");
            msg.sendResponse(reply);
            Log.log(Log.WRITE+Log.DEBUG, ">> return sendResponse");
        }
    }

    //
    // Listener code to accept connections by outside agents 
    //
    /**
     * Peforms the <code>accept</code> calls for new connections to the
     * server socket.
     *
     * <b>Note:  this should not be used by the application.  It must
     * be public to extend the Thread class.</b>
     * @see java.lang.Thread
     */
    public void run()
    {
        Connection newConn;

        setPriority(getPriority() + 2);
        //
        // No support for acting as an ANS in this implementation.
        //
        Log.log(Log.IO+Log.DEBUG, "Agent " + agentName +
		" listener thread started");
        while (true)
        {
            try
            {
                newConn = new Connection(server.accept(), this);
                synchronized (connCache)
                {
                    Log.log(Log.IO+Log.DEBUG,
                            "Add new connection to cache from " +
                            newConn.getHost());
                    connCache.addElement(newConn);
                }
            }
            catch (java.io.IOException e)
            {
                Log.log(Log.ERROR, "Accept Exception: " + e.getMessage());
            }
            catch (InvalidKQMLOpenException e)
            {
                Log.log(Log.ERROR, "Connection Open Exception: " +
			e.getMessage());
            }
        }
    }

    /**
     * Gets the agent name of an agent instance.  Used principally
     * by the KRIL.
     * @return the agent's name
     */
    public String getAgentName()
    {
        return (agentName);
    }

    /**
     * Gets the active state name of an agent instance.  The active state
     * will be false if the agent has been shutdown.
     * @return the agent's active state
     */
    public boolean agentActive()
    {
        return (agentState);
    }

    /**
     * Gets the default sendAndAwaitReply timeout value for the agent
     * instance.
     * @return the agent's default reply timeout value
     */
    public int getTimeout()
    {
        return (timeout);
    }

    /**
     * Sets the default sendAndAwaitReply timeout value for the agent
     * instance.
     * @parameter timeout the new value for agent's default reply timeout value
     */
    public void setTimeout(int timeout)
    {
        this.timeout = timeout;
    }

    //
    // KQML "Timing" functions
    //
    /**
     * Disables the agent's timing enabled state.
     */
    public void disableTiming()
    {
        timingEnabled = false;
        timingTable = new Hashtable();
    }

    /**
     * Enables the agent's timing enabled state.  The state is
     * initially disabled (that is false).
     */
    public void enableTiming()
    {
        timingEnabled = true;
    }

    /**
     * Gets the agent instance's timing enabled state.
     * @return the agent's timing enabled state.
     */
    public boolean timingEnabled() { return (timingEnabled); }


    //
    // KQML "Logging" functions
    //
    /**
     * Disables the agent's logging enabled state.
     */
    public void disableLogging()
    {
        loggingEnabled = false;
        loggingTable = new Hashtable();
    }

    /**
     * Enables the agent's logging enabled state.  The state is
     * initially disabled (that is false).
     */
    public void enableLogging()
    {
        loggingEnabled = true;
    }

    /**
     * Gets the agent instance's logging enabled state.
     * @return the agent's logging enabled state.
     */
    public boolean loggingEnabled() { return (loggingEnabled); }

    /**
     * Sends a message to the agent designated in the receiver parameter.
     * @param perf     message performative
     * @param content  message content
     * @param receiver message receiver agent
     * @exception kqml.InvalidKQMLMsgException when the performative is null
     * @exception kqml.KqmlIOException usally due to remote socket closure
     * or network failure.
     */
    public void sendMessage(String perf, String content, String receiver)
                                throws kqml.InvalidKQMLMsgException,
                                       kqml.KqmlIOException
    {
        new Message(perf, content, receiver, this).send();
    }

    /**
     * Sends a message to the agent designated in the receiver parameter with
     * an array of strings concatenated in the <code>:content</code> field.
     * @param perf     message performative
     * @param array    array of strings to be concatenated in the message's
     * <code>:content</code> field
     * @param receiver message receiver agent
     * @exception kqml.InvalidKQMLMsgException when the performative is null
     * @exception kqml.KqmlIOException usally due to remote socket closure
     * or network failure.
     */
    public void sendMessage(String perf, String[] array, String receiver)
                                throws kqml.InvalidKQMLMsgException,
                                       kqml.KqmlIOException
    {
        new Message(perf, array, receiver, this).send();
    }

    /**
     * Sends a message to the agent designated in the receiver parameter with
     * a vector of strings concatenated in the <code>:content</code> field.
     * @param perf     message performative
     * @param vector   vector of strings to be concatenated into the message's
     * <code>:content</code> field
     * @param receiver message receiver agent
     * @exception kqml.InvalidKQMLMsgException when the performative is null
     * @exception kqml.KqmlIOException usally due to remote socket closure
     * or network failure.
     */
    public void sendMessage(String perf, Vector vector, String receiver)
                                throws kqml.InvalidKQMLMsgException,
                                       kqml.KqmlIOException
    {
        new Message(perf, vector, receiver, this).send();
    }

    /**
     * Sends a message to the agent designated in the receiver parameter with
     * an integer converted to string in the <code>:content</code> field.
     * @param perf     message performative
     * @param i        message integer value converted to a string
     * @param receiver message receiver agent
     * @exception kqml.InvalidKQMLMsgException when the performative is null
     * @exception kqml.KqmlIOException usally due to remote socket closure
     * or network failure.
     */
    public void sendMessage(String perf, int i, String receiver)
                                throws kqml.InvalidKQMLMsgException,
                                       kqml.KqmlIOException
    {
        new Message(perf, i, receiver, this).send();
    }

    /**
     * Sends a message to the agent designated in the receiver parameter with
     * an array of integers converted to string in the <code>:content</code>
     * field.
     * @param perf     message performative
     * @param i        message integer array converted to a string
     * @param receiver message receiver agent
     * @exception kqml.InvalidKQMLMsgException when the performative is null
     * @exception kqml.KqmlIOException usally due to remote socket closure
     * or network failure.
     */
    public void sendMessage(String perf, int[] i, String receiver)
                                throws kqml.InvalidKQMLMsgException,
                                       kqml.KqmlIOException
    {
        new Message(perf, i, receiver, this).send();
    }

    /**
     * Sends a message to the agent designated in the receiver parameter with
     * an long converted to string in the <code>:content</code> field.
     * @param perf     message performative
     * @param l        message long value converted to a string
     * @param receiver message receiver agent
     * @exception kqml.InvalidKQMLMsgException when the performative is null
     * @exception kqml.KqmlIOException usally due to remote socket closure
     * or network failure.
     * @exception kqml.InvalidLongInputException the long value needs 
     * more than 32 bits to be represented.
     */
    public void sendMessage(String perf, long l, String receiver)
                                throws kqml.InvalidKQMLMsgException,
                                       kqml.InvalidLongInputException,
                                       kqml.KqmlIOException
    {
        new Message(perf, l, receiver, this).send();
    }

    /**
     * Sends a message to the agent designated in the receiver parameter with
     * an array of longs converted to string in the <code>:content</code>
     * field.
     * @param perf     message performative
     * @param l        message long array converted to a string
     * @param receiver message receiver agent
     * @exception kqml.InvalidKQMLMsgException when the performative is null
     * @exception kqml.KqmlIOException usally due to remote socket closure
     * or network failure.
     * @exception kqml.InvalidLongInputException one of the long values needs 
     * more than 32 bits to be represented.
     */
    public void sendMessage(String perf, long[] l, String receiver)
                                throws kqml.InvalidKQMLMsgException,
                                       kqml.InvalidLongInputException,
                                       kqml.KqmlIOException
    {
        new Message(perf, l, receiver, this).send();
    }

    /**
     * Sends a message to the agent designated in the receiver parameter
     * and returns the expected reply.
     * @param perf     message performative
     * @param content  message content
     * @param receiver message receiver agent
     * @exception kqml.InvalidKQMLMsgException when the performative is null
     * @exception kqml.InvalidKQMLTokenException when the reply is malformed
     * @exception kqml.KQMLMsgTimeout when the reply fails to arrive on time
     * @exception kqml.KqmlIOException usally due to remote socket closure
     * or network failure.
     */
    public Message sendMsgAndAwaitReply(String perf, String content,
                                        String receiver)
                                throws kqml.InvalidKQMLMsgException,
                                       kqml.InvalidKQMLTokenException,
                                       kqml.KQMLMsgTimeout,
                                       kqml.KqmlIOException
    {
        return (new Message(perf, content, receiver, this).sendAndAwaitReply());
    }

    /**
     * Sends a message to the agent designated in the receiver parameter
     * and returns the expected reply, with the timeout period specified.
     * @param perf     msg performative
     * @param content  msg content
     * @param receiver msg receiver agent
     * @param timeout  time to wait for a reply, in milliseconds, before a
     * timeout
     * @exception kqml.InvalidKQMLMsgException when the performative is null
     * @exception kqml.InvalidKQMLTokenException when the reply is malformed
     * @exception kqml.KQMLMsgTimeout when the reply fails to arrive on time
     * @exception kqml.KqmlIOException usally due to remote socket closure
     * or network failure.
     */
    public Message sendMsgAndAwaitReply(String perf, String content,
                                        String receiver, long timeout)
                                throws kqml.InvalidKQMLMsgException,
                                       kqml.InvalidKQMLTokenException,
                                       kqml.KQMLMsgTimeout,
                                       kqml.KqmlIOException
    {
        return (new Message(perf, content, receiver, this).
                                    sendAndAwaitReply(timeout));
    }

    /**
     * Sends a message to the agent designated in the receiver parameter with
     * a vector of strings concatenated in the <code>:content</code> field
     * and returns the expected reply.
     * @param perf     message performative
     * @param vector   message vector of strings to be concatenated into
     * the message's <code>:content</code> field
     * @param receiver message receiver agent
     * @exception kqml.InvalidKQMLMsgException when the performative is null
     * @exception kqml.InvalidKQMLTokenException when the reply is malformed
     * @exception kqml.KQMLMsgTimeout when the reply fails to arrive on time
     * @exception kqml.KqmlIOException usally due to remote socket closure
     * or network failure.
     */
    public Message sendMsgAndAwaitReply(String perf, Vector vector,
                                        String receiver)
                                throws kqml.InvalidKQMLMsgException,
                                       kqml.InvalidKQMLTokenException,
                                       kqml.KQMLMsgTimeout,
                                       kqml.KqmlIOException
    {
        return (new Message(perf, vector, receiver, this).sendAndAwaitReply());
    }

    /**
     * Sends a message to the agent designated in the receiver parameter with
     * a vector of strings concatenated in the <code>:content</code> field
     * and returns the expected reply.
     * @param perf     message performative
     * @param vector   message vector of strings to be concatenated into
     * the message's <code>:content</code> field
     * @param receiver message receiver agent
     * @param timeout  time to wait for a reply, in milliseconds, before a
     * timeout
     * @exception kqml.InvalidKQMLMsgException when the performative is null
     * @exception kqml.InvalidKQMLTokenException when the reply is malformed
     * @exception kqml.KQMLMsgTimeout when the reply fails to arrive on time
     * @exception kqml.KqmlIOException usally due to remote socket closure
     * or network failure.
     */
    public Message sendMsgAndAwaitReply(String perf, Vector vector,
                                        String receiver, long timeout)
                                throws kqml.InvalidKQMLMsgException,
                                       kqml.InvalidKQMLTokenException,
                                       kqml.KQMLMsgTimeout,
                                       kqml.KqmlIOException
    {
        return (new Message(perf, vector, receiver, this).
                                                sendAndAwaitReply(timeout));
    }
    
    protected void setKmsgTime(String id, int type, Date start, Date stop)
    {
        if (timingEnabled && (id != null))
        {
            TimingRecord    tr;

            try
            {
                if ((tr = (TimingRecord)timingTable.get(id)) == null)
                {
                    timingTable.put(id,
				    new TimingRecord(id, type, start,stop));
                    Log.log(Log.WRITE+Log.DEBUG,
                            "setKmsgTime: setup timing record\n     " + id);
                }
                else
                {
                    Log.log(Log.WRITE+Log.DEBUG,
                            "setKmsgTime: complete timing record\n     " + id);
                    //
                    // completeTimingRecord will also send the record to
                    // the Timings agent
                    //
                    tr.completeTimingRecord(type, start, stop);
                    timingTable.remove(id);
                    admin.putWork(tr);
                }
            }
            catch (TimingRecordException e)
            {
                Log.log(Log.ERROR, "setKmsgTime: Internal error " +
			e.getMessage());
            }
        }
    }

    /**
     * Gets the KRIL version string from the Version class.
     * @return the KRIL version string.
     */
    public String getKrilVersion()
    {
        return (Version.krilVersion);
    }

    //
    // Thread msg queue get and put functions
    //
    protected Object pendingMsgGetThread(Message msg)
    {
        Object  obj;
	String  kqmlMsgId = msg.getKqmlMsgId();
	
        synchronized (pendingMsgs)
        {
	  obj = pendingMsgs.get(kqmlMsgId);
	  pendingMsgs.remove(kqmlMsgId);
        }
        return (obj);
    }

    protected void pendingMsgPutThread(Message msg, Object obj)
    {
        synchronized (pendingMsgs)
        {
	  pendingMsgs.put(msg.getKqmlMsgId(), obj);
        }
    }

  // this pending message has timed out, so replace it's thread reference
  // in the pendingMsgs hashtable with an instance of a TimedOutMessage
  protected void pendingMsgTimedOut(Object key)
  {
    TimedOutMessage timedOutMsg = new TimedOutMessage();
    
    Log.log(Log.DEBUG, "Pending Msg '" + key + "' has Timed Out");
    synchronized (pendingMsgs)
      {
	pendingMsgs.put(key, timedOutMsg);
      }
  }
  
    //
    // If a message is found, determine if a specific handler thread
    // is waiting for this particular message  (identified by the
    // "kqml-msg-id" field).  If a particular thread is waiting, put the
    // message in that thread's message queue.  Otherwise, put the message
    // in the unsolicited message queue.
    //
    protected void scanCacheForMsgs()
    {
        int         cacheSize;
        Connection  conn;

        synchronized (connCache)
        {
            cacheSize = connCache.size();
        }

        if (lastCacheSize != cacheSize)
        {
            Log.log(Log.DEBUG, "scanCache: Cache Size is " + cacheSize);
            lastCacheSize = cacheSize;
        }

        for (int i = 0; i < cacheSize; ++i) 
        {
            synchronized (connCache)
            {
                try
                { 
                    conn = (Connection)connCache.elementAt(i);
                }
                catch (ArrayIndexOutOfBoundsException e)
                {
                    conn = null;
                }
            }
            if (conn != null)
                processConn(conn);
        }

        checkTimeouts();
    }

    private void processConn(Connection conn)
    {
      Message     msg;
      Object      obj;
      
      try
        {
	  if ((msg = conn.asyncGet()) != null)
            {
	      //
	      // Determine if this message is destined for a particular
	      // thread
	      //
	      if ((obj = pendingMsgGetThread(msg)) == null)
                {
		  //
		  // This message is unsolicited, so put in the
		  // unsolicited message queue and wake up any idle
		  // msgThread to handle the message.
		  //
		  unsolicitedQueue.putMsg(msg, msgThreads);
                }
	      else
                {
		  String threadName = null;
		  //
		  // This message is destined for a particular thread...
		  
		  // ...however, the wait time for this reply already
		  //    expired, so throw it away.
		  if (obj instanceof TimedOutMessage)
		    {
		      Log.log(Log.ERROR, "Reply for Timed-Out Message " +
			      msg.getKqmlMsgId());
		    }
		  // ...and the thread is still waiting for it, so put the
		  //    message in this thread's message queue
		  //
		  else if (obj instanceof KqmlDispatch)
                    {
		      ((KqmlDispatch)obj).putExpectedMsg(msg);
		      threadName = ((KqmlDispatch)obj).getThread().getName();
                    }
		  else if (obj instanceof KqmlRegistration)
                    {
		      ((KqmlRegistration)obj).putExpectedMsg(msg);
		      threadName =
			((KqmlRegistration)obj).getThread().getName();
                    }
		  else if (obj instanceof KqmlAdmin)
                    {
		      ((KqmlAdmin)obj).putExpectedMsg(msg);
		      threadName = ((KqmlAdmin)obj).getThread().getName();
                    }
		  Log.log(Log.READ+Log.DEBUG, "scanCache: msg from " +
			  conn.getRemoteAgent() +
			  "\n     is solicited for " + threadName);
                }
            }
        }
      
      //
      // If reading on this socket results in an EOFException
      // remove this Connection from the cache.
      //
      catch (java.io.EOFException e)
        {
	  Log.log(Log.ERROR, "EOF on Socket " + conn.getHost() + " " +
		  e.getMessage());
	  cleanupConnection(conn);
        }
      catch (java.io.IOException e)
        {
	  
	  Log.log(Log.ERROR, "IO Error on Socket " + conn.getHost() +
		  " " + e.getMessage());
	  cleanupConnection(conn);
        }
    }
  
    private void checkTimeouts()
    {
      long   now     = System.currentTimeMillis();
      Object obj     = null;
      Object key     = null;
      
      for (Enumeration e = pendingMsgs.keys(); e.hasMoreElements(); )
        {
	  key = e.nextElement();
	  obj = pendingMsgs.get(key);

	  if (obj instanceof KqmlDispatch)
            {
	      if (((KqmlDispatch)obj).getTimeout() < now)
		{
		  pendingMsgTimedOut(key);

		  ((Thread)obj).resume();
		}
            }
	  else if (obj instanceof KqmlRegistration)
            {
	      if (((KqmlRegistration)obj).getTimeout() < now)
		{
		  pendingMsgTimedOut(key);

		  ((KqmlRegistration)obj).getThread().resume();
		}
            }
	  else if (obj instanceof KqmlAdmin)
            {
	      if (((KqmlAdmin)obj).getTimeout() < now)
		{
		  pendingMsgTimedOut(key);

		  ((KqmlAdmin)obj).resume();
		}
            }
	  else if (obj instanceof TimedOutMessage)
            {
	      Log.log(Log.DEBUG, "        value: TimedOutMessage");
            }
        }
    }

    //
    // Determine the ANS host id ... currently via the java System
    // Property mechanism.
    //
    private static String getAnsHost() throws KqmlAgentRegistrationFailure
    {
        String defaultHost = null;
        String rval;

        try
        {
            defaultHost = InetAddress.getLocalHost().getHostAddress();
        }
        catch (UnknownHostException e) { }
        try
        {
            if ((rval=System.getProperty("ANS_HOST")) == null)
                rval = defaultHost;
        }
        catch (SecurityException e)
        {
            rval = defaultHost;
        }

        if (rval == null)
            throw new KqmlAgentRegistrationFailure(
                        "getAnsHost can't determine local host id and" +
                        " ANS_HOST property not set");
        return (rval);
    }

    //
    // Determine the KQML TIMEOUT value ... currently via the java System
    // Property mechanism.
    //
    // Later, supplement by checking a configuration file
    //
    private int getKqmlTimeout()
    {
        String property;
        int rval;

        try
        {
            if ((property=System.getProperty("KQML_TIMEOUT")) == null)
                rval = DEFAULT_TIMEOUT;
            else
                rval = Integer.parseInt(property) * MILLISECONDS_PER_SEC;
        }
        catch (SecurityException e)
        {
            rval = DEFAULT_TIMEOUT;
        }
        catch (NumberFormatException e)
        {
            rval = DEFAULT_TIMEOUT;
            Log.log(Log.ERROR, "KQML_TIMEOUT property setting not numeric" +
		    e.getMessage());
            System.exit(1);
        }
        return (rval);
    }
    
    private synchronized void cleanupConnection(Connection conn)
    {
        String name;

        if ((name = conn.getRemoteAgent()) == null)
        {
            Log.log(Log.IO+Log.DEBUG, "cleanup Connection `" + conn + "'");
        }
        else
        {
            agentLookup.remove(name);
            Log.log(Log.IO+Log.DEBUG, "cleanup Connection to `" + name + "'");
        }
        conn.close();
        connCache.removeElement(conn);
    }

    protected Connection lookupAgent(String agentName)
                                            throws InvalidKQMLMsgException
    {
        Connection rval = null;

        synchronized (agentLookup)
        {
            rval = (Connection)agentLookup.get(agentName);
        }

        if (rval == null)
        {
            rval = ansLookupAndConnect(agentName);
            if (rval == null)
                throw new InvalidKQMLMsgException (DESTINATION_AGENT_UNKNOWN
                    + agentName + "'");
            else
                addAgentToLookup(rval.getRemoteAgent(), rval);
        }
        return (rval);
    }

    protected void addAgentToLookup(String agentName, Connection conn)
    {
        Connection prevConn;

        synchronized (agentLookup)
        {
            prevConn = (Connection)agentLookup.put(agentName, conn);
        }
        connCache.addElement(conn);

        //
        // If there was a previous connection with this agent name
        // clean it up
        //if (prevConn != null)
        //  cleanupConnection(prevConn);
    }

    private Connection ansLookupAndConnect(String agentName)
    {
        Connection rval = null;
        Message lookUp;

        //
        // Create the "ask-one" message, send it to the ANS, get
        // the response back, and verify a successful registration
        //
        try
        {
            String  contents;
            Message response;

            lookUp = new Message(ans, ASK_ONE, agentName, ansAgent);

            Log.log(Log.WRITE+Log.DEBUG, "ANS Lookup " + ASK_ONE + " " +
		    agentName);
            response      = lookUp.sendAndAwaitReply(ANS_LOOKUP_TIMEOUT);
            contents      = response.getContents();

            if (response.perfIsReply())
            {
                rval = new Connection(new AgentAddress(contents), this);
                if (notRegistered.get(agentName) != null)
                    notRegistered.remove(agentName);

                Log.log(Log.READ+Log.DEBUG, "Agent " + agentName + " AT " +
			contents);
            }
            else if (response.perfIsSorry())
            {
                //
                // This bracketing logic prevents repeated reporting
                // of the ANS lookup failure.
                //
                if (notRegistered.get(agentName) == null)
                {
                    Log.log(Log.ERROR, "Agent " + agentName +
                            " NOT REGISTERED with ANS " + ansAgent);
                    notRegistered.put(agentName, agentName);
                }
            }
            else
            {
                if (notRegistered.get(agentName) == null)
                {
                    Log.log(Log.ERROR, "Unexpected ANS response from " +
                            ansAgent + " for agent " + agentName + " lookup ");
                    Log.log(Log.ERROR, "perf: `" + response.getPerformative() +
                            "' content: `" + contents + "'");
                }
            }
        }
        catch (KqmlException e)
        {
            if (notRegistered.get(agentName) == null)
            {
                Log.log(Log.ERROR, "KQML exception on ANS lookup of " +
			agentName + ": " + e.getMessage());
                notRegistered.put(agentName, agentName);
            }
        }
        return (rval);
    }

    //
    // Has to called (and executed) from the thread in question
    // -- thus KqmlAdmin, KqmlDispatch and KqmlRegistration constructors
    //    MUST call this method in order for sendWithReply to be able
    //    report the result back to the correct thread
    //
    protected void addLookupThread(Object o)
    {
        threadLookup.put(Thread.currentThread(), o);
        Log.log(Log.THREAD, "adding Thread " +
		Thread.currentThread().getName());
    }

    protected Object lookupThread()
    {
        Object rval = threadLookup.get(Thread.currentThread());

        if (rval instanceof KqmlDispatch)
            Log.log(Log.THREAD, "lookupThread returning " +
                    ((KqmlDispatch)rval).getThread().getName());
        else if (rval instanceof KqmlRegistration)
            Log.log(Log.THREAD, "lookupThread returning " +
                    ((KqmlRegistration)rval).getThread().getName());
        else if (rval instanceof KqmlAdmin)
            Log.log(Log.THREAD, "lookupThread returning " +
                    ((KqmlAdmin)rval).getThread().getName());
        return (rval);
    }
}
