/*
** Restricted Rights Legend
**
** Use, duplication or disclosure by the government is subject to restrictions
** as set forth in subparagraph (c)(1)(ii) of the Rights in Technical Data and
** Computer Software clause at DFARS 252.227-7013.
**
** Copyright (c) 1997 Lockheed Martin C2 Integration Systems
**
**
** File:         KqmlDispatch.java
**
** Description:  Contains the KqmlDispatch class that is internal to
**               the KRIL.
**
*/
package kqml;

/**
 * A class internal to the KRIL.
 *
 * This class is run as a thread and used to "dispatch" messages
 * and invoking the agent's "dispatch" method to lookup the handler
 * method (via a Handler object) to execute for a particular KQML
 * performative.  It also provides a distinct "run" method from the
 * "Listener" run method provided in the KqmlAgent class.
 *
 * @author Peter Stevens
 * @version $Revision: 1.8 $
 * @see KqmlAgent
 */

class KqmlDispatch extends Thread
{
    private KqmlAgent         agent       = null;
    private boolean           busy        = false;
    private MsgQueue          msgQueue    = null;
    private UnsolicitedQueue  unsolicited = null;
    private long              timeout     = 0;
    private static int        number      = 0;
    private static final long PENDING_MSGS_TIME_INTERVAL = 100000;

    // Constructor
    KqmlDispatch(KqmlAgent agent, UnsolicitedQueue unsolicited)
    {
        super("Dispatch" + number++);
        this.agent       = agent;
        msgQueue         = new MsgQueue(this);
        this.unsolicited = unsolicited;
    }

    public void run()
    {
        Log.log(Log.THREAD, "Agent " + agent.getAgentName() +
            " starting dispatch thread `" + this.getName() + "'");
        agent.addLookupThread(this);    // for sendWithReply operations

        while (true)
        {
            // Get a message from this thread's unsolicited message queue
            Message msg = unsolicited.getMsg();

            if (msg != null)
            {
	      // Process it by using the agent's dispatch method
	      synchronized (this)
		{
		  busy = true;
		}
	      agent.dispatch(msg);
	      synchronized (this)
                {
		  busy = false;
                }
            }
        }
    }

    protected void makePending(Message msg)
    {
        //
        // Register the thread as waiting for this message
        //
        agent.pendingMsgPutThread(msg, this);

	// Since this message is now on the pendingMsgs queue, it can be
	// accessed by the checkTimeouts() method of the KqmlAgent object
	// which will check to see if it has timed out.  So, we must reset
	// the timeout value for this thread here.  Since we don't want to
	// set that value for real until getExpectedMsg() is called, set it
	// to something big enough that it will allow for the different 
	// threads to catch up to each other.
	this.timeout = System.currentTimeMillis() + PENDING_MSGS_TIME_INTERVAL;

        Log.log(Log.IO+Log.DEBUG, "KqmlDispatch:makePending msgId\n     `" +
                msg.getKqmlMsgId() + "' for thread `" + this.getName() +
                "'\n     perf `" + msg.getPerformative() +  "' from `" +
                msg.getReceiver() + "'");
    }

    protected Message getExpectedMsg(long timeout)
    {
        //
        // Save the timeout value for this thread computed as time in the
        // future
        //
        this.timeout = System.currentTimeMillis() + timeout;
	Log.log(Log.DEBUG, "getExpectedMsg: set timeout=" + this.timeout +
		" for thread `" + this.getName() + "'");

        //
        // Block this thread, waiting for the message
        //
        return (msgQueue.getExpectedMsg());
    }

    protected void putExpectedMsg(Message msg)
    {
        msgQueue.putExpectedMsg (msg);
    }
    
    protected long getTimeout()
    {
        return (timeout);
    }
 
    protected Thread getThread()
    {
        return ((Thread)this);
    }

    protected boolean isBusy()
    {
        boolean rval;

        synchronized (this)
        {
            rval = busy;
        }
        return (rval);
    }
}
