/*
** Restricted Rights Legend
**
** Use, duplication or disclosure by the government is subject to restrictions
** as set forth in subparagraph (c)(1)(ii) of the Rights in Technical Data and
** Computer Software clause at DFARS 252.227-7013.
**
** Copyright (c) 1997 Lockheed Martin C2 Integration Systems
**
**
** File:        Message.java
**
** Description: Contains the Message and Token classes.  They have the
**              following "public" methods:
**
**              Message:  constructors to generate a message
**                        and methods to send and sendAndAwaitReply 
**                        "set" the performative, reply-with, content
**                        kril-wrapper, receiver, sender, kqml-msg-id
**                        and "other" keyword value pairs.  There are
**                        corresponding "get" methods for these fields.
**                        The "readMessage" function, is the other key
**                        method (but not to the "public") used to parse
**                        incoming messages.
**
**              Token:    constructor to generate a message "token"
**                        to for assistance parsing keyword-value pairs
**                        mentioned above.
*/

package kqml;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.TimeZone;
import java.util.Vector;

//
// Objects of type "message" are used to send messages to external
// KQML agents and to receive messages from them.
//

/**
 * A class to construct and send an outgoing message and parse an
 * incoming message.
 *
 * @author Michael Horton
 * @author Peter Stevens (current)
 * @version $Revision: 1.7 $
 * @see Handler
 * @see KQMLMsgTimeout
 * @see KqmlAgent
 * @see TimedOutMessage
 * @see kqml.InvalidKQMLMsgException 
 * @see kqml.InvalidKQMLTokenException
 * @see kqml.KqmlIOException
 * @see kqml.Exception
 */
public class Message implements KqmlKeywords
{
  private KqmlAgent  agent         = null;
  private Hashtable  keywordValues = new Hashtable();
  private Connection connect       = null;
  private String     performative  = "";
  
  private static final boolean DEBUG = false;
  
  private static final String START_TOKEN_STRING = "!@#$%";
  
  private static final String PERF_ERROR = "performative name cannot be null";
  
  private static final String doubleIndentedLine = "\n          ";
  private static final String indentedLine       = "\n     ";
  private static final int doubleIndentedLineLen =
  doubleIndentedLine.length();
  private static final int indentedLineLen       = indentedLine.length();
  
  private static String           agentId;
  private static SimpleDateFormat formatter;
  private static int              generatedMsgNo = 0;
  
  static
  {
    try
      {
	InetAddress addr = InetAddress.getLocalHost();
	
	agentId = addr.getHostName() + "+";
	if (agentId == null)
	  agentId = addr.toString() + "+";
	agentId += System.currentTimeMillis() + "+";
      }
    catch (UnknownHostException e)
      {
	agentId = "UnknownHost+" + System.currentTimeMillis() + "+";
      }
    formatter = new SimpleDateFormat("yD.HH:mm:ss.SSS");
    formatter.setTimeZone(TimeZone.getDefault());
  }
  
  //
  // Constructors
  // - this set for directing message/replies to a specific connection
  //   and are used "internally" in the KRIL
  //
  protected Message (Connection connect)
  {
    keywordValues.put(SENDER, connect.getAgentName());
    this.connect        = connect;
    this.agent          = connect.getAgent();
  }
  
  protected Message (Connection connect,
		     String     performative,
		     String     contents,
		     String     receiver) throws kqml.InvalidKQMLMsgException
  {
    this.performative   = performative;
    this.connect        = connect;
    this.agent          = connect.getAgent();
    keywordValues.put(CONTENT,      contents);
    keywordValues.put(RECEIVER,     receiver);
    keywordValues.put(SENDER,       connect.getAgentName());
    keywordValues.put(KRIL_WRAPPER,
		      new KrilWrapper(contents).getKrilWrapper());
  }
  
  //
  // - This set for messages to a destination agent (a connection hasn't
  //   necessary been established but might be in the lookup cache)
  // - these are generally for "users"
  //
  
  /**
   * Returns a message instance with the performative, content, sender
   * and receiver agent filled in.  A kril-wrapper will automatically
   * be generated.
   *
   * @param performative   the Message's performative
   * @param contents       the value of the content field
   * @param receive        the name of the destination agent
   * @param agent          reference to the agent that will send the message
   * @exception InvalidKQMLMsgException  if the either of the performative
   *                                     or receiver strings are null
   * @see KqmlAgent
   */
  public Message (String    performative,
		  String    contents,
		  String    receiver,
		  KqmlAgent agent) throws InvalidKQMLMsgException
  {
    this.performative = performative;
    this.agent        = agent;
    keywordValues.put(CONTENT,      contents);
    keywordValues.put(RECEIVER,     receiver);
    keywordValues.put(SENDER,       agent.getAgentName());
    keywordValues.put(KRIL_WRAPPER,
		      new KrilWrapper(contents).getKrilWrapper());
  }
  
  /**
   * Returns a message instance with the performative, content, sender
   * and receiver agent filled in.  A kril-wrapper will automatically
   * be generated.
   *
   * @param performative   the Message's performative
   * @param vec            value assumed to a vector of strings, concatenated
   *                       to each other and fills in the message's content
   *                       field
   * @param receive        the name of the destination agent
   * @param agent          reference to the agent that will send the message
   * @exception InvalidKQMLMsgException  if the either of the performative
   *                                     or receiver strings are null
   * @see KqmlAgent
   */
  public Message (String    performative,
		  Vector    vec,
		  String    receiver,
		  KqmlAgent agent)  throws kqml.InvalidKQMLMsgException
  {
    this(performative, "", receiver, agent);
    
    KrilWrapper wrap = new KrilWrapper(vec);
    keywordValues.put(CONTENT, wrap.getContent());
    keywordValues.put(KRIL_WRAPPER, wrap.getKrilWrapper());
  }
  
  /**
   * Returns a message instance with the performative, content, sender
   * and receiver agent filled in.  A kril-wrapper will automatically
   * be generated.
   *
   * @param performative  the Message's performative
   * @param array         an array of strings, concatenated to each other
   *                      and fills in the message's content field
   * @param receiver      the name of the destination agent
   * @param agent         reference to the agent that will send the message
   * @exception InvalidKQMLMsgException  if the either of the performative
   *                                     or receiver strings are null
   * @see KqmlAgent
   */
  public Message (String    performative,
		  String[]  array,
		  String    receiver,
		  KqmlAgent agent)  throws kqml.InvalidKQMLMsgException
  {
    this (performative, "", receiver, agent);
    
    KrilWrapper wrap = new KrilWrapper(array);
    keywordValues.put(CONTENT,      wrap.getContent());
    keywordValues.put(KRIL_WRAPPER, wrap.getKrilWrapper());
  }
  
  /**
   * Returns a message instance with the performative, content, sender
   * and receiver agent filled in.  A kril-wrapper will automatically
   * be generated.
   *
   * @param performative   the Message's performative
   * @param l              a long value converted to a string and put in the
   *                       message's content field
   * @param receiver       the name of the destination agent
   * @param agent          reference to the agent that will send the message
   * @exception InvalidKQMLMsgException    if the either of the performative
   *                                       or receiver strings are null
   * @exception InvalidLongInputException  if the long input needs more
   *                                       than 32 bits to be represented
   * @see KqmlAgent
   */
  public Message (String    performative,
		  long      l,
		  String    receiver,
		  KqmlAgent agent)
       throws kqml.InvalidKQMLMsgException,
		    kqml.InvalidLongInputException
  {
    this (performative, "", receiver, agent);
    
    KrilWrapper wrap = new KrilWrapper(l);
    keywordValues.put(CONTENT,      wrap.getContent());
    keywordValues.put(KRIL_WRAPPER, wrap.getKrilWrapper());
  }
  
  /**
   * Returns a message instance with the performative, content, sender
   * and receiver agent filled in.  A kril-wrapper will automatically
   * be generated.
   *
   * @param performative   the Message's performative
   * @param                array an array of longs, each converted to String,
   *                       concatenated to each other and fills in the
   *                       message's content field
   * @param receiver       the name of the destination agent
   * @param agent          reference to the agent that will send the message
   * @exception InvalidKQMLMsgException    if the either of the performative
   *                                       or receiver strings are null
   * @exception InvalidLongInputException  if the long input needs more
   *                                       than 32 bits to be represented
   * @see KqmlAgent
   */
  public Message (String    performative,
		  long[]    array,
		  String    receiver,
		  KqmlAgent agent)
       throws kqml.InvalidKQMLMsgException,
		    kqml.InvalidLongInputException
  {
    this (performative, "", receiver, agent);
    
    KrilWrapper wrap = new KrilWrapper(array);
    keywordValues.put(CONTENT,      wrap.getContent());
    keywordValues.put(KRIL_WRAPPER, wrap.getKrilWrapper());
  }
  
  /**
   * Returns a message instance with the performative, content, sender
   * and receiver agent filled in.  A kril-wrapper will automatically
   * be generated.
   *
   * @param performative   the Message's performative
   * @param i              an integer value converted to a string and
   *                       put in the message's content field
   * @param receiver       the name of the destination agent
   * @param agent          reference to the agent that will send the message
   * @exception InvalidKQMLMsgException  if the either of the performative
   *                                     or receiver strings are null
   * @see KqmlAgent
   */
  public Message (String    performative,
		  int       i,
		  String    receiver,
		  KqmlAgent agent)  throws kqml.InvalidKQMLMsgException
  {
    this (performative, "", receiver, agent);
    
    KrilWrapper wrap = new KrilWrapper(i);
    keywordValues.put(CONTENT,      wrap.getContent());
    keywordValues.put(KRIL_WRAPPER, wrap.getKrilWrapper());
  }
  
  /**
   * Returns a message instance with the performative, content, sender
   * and receiver agent filled in.  A kril-wrapper will automatically
   * be generated.
   *
   * @param performative   the Message's performative
   * @param array          an array of integers, each converted to String,
   *                       concatenated to each other and fills in the
   *                       message's content field
   * @param receiver       the name of the destination agent
   * @param agent          reference to the agent that will send the message
   * @exception InvalidKQMLMsgException  if the either of the performative
   *                                     or receiver strings are null
   * @see KqmlAgent
   */
  public Message (String    performative,
		  int[]     array,
		  String    receiver,
		  KqmlAgent agent)  throws kqml.InvalidKQMLMsgException
  {
    this (performative, "", receiver, agent);
    
    KrilWrapper wrap = new KrilWrapper(array);
    keywordValues.put(CONTENT,      wrap.getContent());
    keywordValues.put(KRIL_WRAPPER, wrap.getKrilWrapper());
  }
  
  // Methods
  //
  // "equals" and "hashCode" should return the same value for the
  // same identifier.
  //
  // NOTE: the current methods only check for matching kqmlMsgId's
  //       this IS REQUIRED because multi-threaded operation of an
  //       agent can result in than one outstanding message from
  //       the same originating agent
  //
  /**
   * Returns the truth of whether two Message's have the same
   * kqml-msg-id.  This is present for insertion & search of
   * Message's in java Hashtable's.  Principally used internally.
   * @param obj reference to the Message whose kqml-msg-id is to
   *            to be compared.
   * @return equality of two message's kqml-msg-id fields.
   * @see java.util.Hashtable
   */
  public boolean equals(Object obj)
  {
    String kqmlMsgId  = getKqmlMsgId();
    String otherMsgId = ((Message)obj).getKqmlMsgId();
    
    if (kqmlMsgId == null || otherMsgId == null)
      return false;
    else
      return kqmlMsgId.equals(otherMsgId);
  }
  
  /**
   * Returns the hashCode value of a Message's kqml-msg-id
   * field.  This is present for insertion & search of
   * Message's in java Hashtable's.  Principally used internally.
   * @return hashCode value of the kqmlMsgId or Object's
   * hashCode if the kqmlMsgId field is null.
   * @see java.util.Hashtable
   */
  public int hashCode()
  {
    String kqmlMsgId  = getKqmlMsgId();
    return (kqmlMsgId == null ?  super.hashCode() : kqmlMsgId.hashCode());
  }
  
  /**
   * Returns string output of a Message's performative and keyword value
   * pairs, each on a separate line.  The performative starts in column 1 
   * while the keyword value pairs are each on a new line starting in
   * column 6.  The length of the value is appended to the line.
   * @return string containing the message's performative, each keyword
   * value pair followed by the length of the value field.  When printed,
   * each keyword value pair will be displayed on a separate line.  
   */
  public String toString()
  {
    String name;
    String rval = performative;
    String value;
    
    for (Enumeration keys = keywordValues.keys(); keys.hasMoreElements();)
      {
	name   = (String)keys.nextElement();
	value  = getKeywordValue(name);
	rval  += "\n     " + name + " " + value + " " + value.length();
      }
    return (rval);
  }
  
  /**
   * Returns string output of a Message's performative and keyword value
   * pairs, each on a separate line, but with a limit on the length of
   * a line.  If the line would be too long because the value is too
   * long, the value is truncated and suffixed with a "..".  If the line
   * would be too long with the keyword name, it is truncated, if necessary,       * and the value is displayed on a second line, indented to column 11.
   * The value length is displayed on a third line indented to column 11,
   * in this case.  A truncated keyword name or value is indicated by a
   * trailing "..".
   * @param len the maximum length output for a keyword value pair line.
   * @return string containing the message's performative and each keyword
   * value pair followed by the length of the value field.  When printed,
   * each keyword value pair will be displayed on a separate line.  
   */
  public String toString(int len)
  {
    String line;
    String name;
    String rval = performative;
    String value;
    
    for (Enumeration keys = keywordValues.keys(); keys.hasMoreElements();)
      {
	name   = (String)keys.nextElement();
	value  = getKeywordValue(name);
	line   = "\n     " + name + " " + value + " " + value.length();
	
	if (line.length() <= len)
	  {
	    rval += line;
	  }
	// Can the keyword "name" fit on the line?
	else if ((line.length()-(value.length()+3)) <= len)
	  {
	    String thisLine  = "\n     " + name + " ";
	    int    suffixLen = Integer.toString(value.length()).length() +3;
	    int    excess    = len - (thisLine.length() + suffixLen) + 1;
	    
	    rval += thisLine;
	    if (excess > 0)
	      rval += value.substring(0, excess);
	    rval += ".. " + value.length();
	  }
	// Need to truncate both the keyword name and value here
	else
	  {
	    if ((name.length()-(indentedLineLen+2)) >= len)
	      rval += indentedLine +
		name.substring(0,len-(indentedLineLen+2)+1) + "..";
	    else
	      rval += indentedLine + name;
	    
	    if ((value.length()-(doubleIndentedLineLen+2)) >= len)
	      rval += doubleIndentedLine +
		value.substring(0,len-(doubleIndentedLineLen+2)+1) + "..";
	    else
	      rval += doubleIndentedLine + value;
	    
	    rval += doubleIndentedLine + value.length();
	  }
      }
    return (rval);
  }
  
  protected Connection getConnection()
  {
    return connect;
  }
  
  /**
   * Sets the performative in a message instance.  Usually for a
   * "re-used" message.
   * @param performative the Message's performative
   * @exception kqml.InvalidKQMLMsgException   if the performative is null
   */
  public void setPerformative(String performative)
       throws kqml.InvalidKQMLMsgException
  {
    if (performative != null)
      this.performative = performative;
    else
      throw new InvalidKQMLMsgException(PERF_ERROR);
  }
  
  /**
   * Gets the performative in a message instance.  Usually used to
   * fetch the performative in a reply message.
   * @return the Message's performative
   */
  public String getPerformative()
  {
    return (performative);
  }
  
  /**
   * Checks if the performative in a message instance is the value "reply".
   * Used to check the performative value in a reply message.
   * @return true if the performative is "reply"
   */
  public boolean perfIsReply()
  {
    return (performative.equalsIgnoreCase(REPLY));
  }
  
  /**
   * Checks if the performative in a message instance is the value "sorry".
   * Used to check the performative value in a reply message.
   * @return true if the performative is "sorry"
   */
  public boolean perfIsSorry()
  {
    return (performative.equalsIgnoreCase(SORRY));
  }
  
  /**
   * Checks if the performative in a message instance is the value "error".
   * Used to check the performative value in a reply message.
   * @return true if the performative is "error"
   */
  public boolean perfIsError()
  {
    return (performative.equalsIgnoreCase(ERROR));
  }
  
  /**
   * Gets the content field in a message instance.  Usually used to
   * fetch the content value in a reply message.
   * @return the Message's content field
   */
  public String getContents()
  {
    return ((String)keywordValues.get(CONTENT));
  }
  
  /**
   * Gets the kqml-msg-id field in a message instance.  Usually used to
   * fetch the kqml-msg-id value in a reply message.
   * @return the Message's kqml-msg-id field
   */
  public String getKqmlMsgId()
  {
    String kmsgId = (String) keywordValues.get(KQML_MSG_ID);

    // All messages (those generated internally, and those received
    // from outside) must have a KQML_MSG_ID field.  So if it isn't 
    // present, add it.
    if (kmsgId == null)
      {
	kmsgId = generateKmsgId();
	
	keywordValues.put(KQML_MSG_ID, kmsgId);
      }
	
    return kmsgId;
  }
  
  /**
   * Gets the kril-wrapper field in a message instance.  Usually used
   * to fetch the kril-wrapper in a reply message.
   * @return the Message's kril-wrapper field
   */
  public String getKrilWrapper()
  {
    return ((String)keywordValues.get(KRIL_WRAPPER));
  }
  
  /**
   * Gets the receiver field in a message instance.  Usually used to
   * fetch the receiver value in a reply message.
   * @return the Message's receiver field
   */
  public String getReceiver()
  {
    return ((String)keywordValues.get(RECEIVER));
  }
  
  /**
   * Gets the reply-with field in a message instance.  Usually used to
   * fetch the reply-with value in a reply message.
   * @return the Message's reply-with field
   */
  public String getReplyWith()
  {
    return ((String)keywordValues.get(REPLY_WITH));
  }
  
  /**
   * Gets the sender field in a message instance.  Usually used to
   * fetch the sender value in a reply message.
   * @return the Message's sender field
   */
  public String getSender()
  {
    return ((String)keywordValues.get(SENDER));
  }
  
  /**
   * Gets the value matching the keyword name
   * to the message.
   * @param keyword the new keyword name
   * @return the string value for the keyword, null if the keyword is not
   * present
   */
  public String getKeywordValue(String keyword)
  {
    return ((String)keywordValues.get(keyword));
  }
  
  /**
   * Sets the value of the named keyword.
   * @param keyword the new keyword name
   * @param value the new keyword value
   * @return the previous string value for the keyword, null if the keyword
   * was not present
   */
  public String setKeywordValue(String keyword, String value)
  {
    return ((String)keywordValues.put(keyword, value));
  }
  
  private static final boolean WITH_REPLY    = true;
  private static final boolean WITHOUT_REPLY = false;
  
  /**
   * Sends the message instance to the "receiver" agent.  Agent
   * destination address lookup is performed by <code>send</code>.
   * @exception kqml.KqmlIOException if an I/O exception is generated sending
   *                                 the message.  Usually due to the remote
   *                                 end of the socket closing.
   * @exception kqml.InvalidKQMLMsgException  due to a null performative
   */
  public void send() throws kqml.KqmlIOException, kqml.InvalidKQMLMsgException
  {
    if (connect == null)
      connect = agent.lookupAgent(getReceiver());
    sendText(WITHOUT_REPLY, null);
  }
  
  //
  // non-blocking sendAndAwaitReply
  //
  /**
   * Sends the message instance to the "receiver" agent and <b>waits</b>
   * for a reply <b>Message</b> object. This is also called a "synchronous"
   * send.  Agent destination address lookup is performed by this method.
   *
   * @param timeout   the timeout value desired by the developer
   * @exception kqml.KqmlIOException  if an I/O exception is generated
   *                                  sending the message.  Usually due to
   *                                  the remote end of the socket closing.
   * @exception kqml.InvalidKQMLMsgException    due to a null performative
   * @exception kqml.InvalidKQMLTokenException  due to invalid token in the
   *                                            input message
   * @exception kqml.KQMLMsgTimeout             message reception timeout
   */
  public Message sendAndAwaitReply(long timeout)
       throws kqml.KqmlIOException, kqml.KQMLMsgTimeout,
	 kqml.InvalidKQMLTokenException,
	 kqml.InvalidKQMLMsgException
  {
    Object obj   = agent.lookupThread();
    Message rval = null;
    
    if (connect == null)
      connect = agent.lookupAgent(getReceiver());
    
    sendText(WITH_REPLY, obj);
    
    if (obj instanceof KqmlDispatch)
      Log.log(Log.DEBUG+Log.WRITE,
	      "sendAndAwaitReply(KqmlDispatch): following sendText");
    else if (obj instanceof KqmlRegistration)
      Log.log(Log.DEBUG+Log.WRITE,
	      "sendAndAwaitReply(KqmlRegistration): following sendText");
    else if (obj instanceof KqmlAdmin)
      Log.log(Log.DEBUG+Log.WRITE,
	      "sendAndAwaitReply(KqmlAdmin): following sendText");
    
    if (obj instanceof KqmlDispatch)
      rval = ((KqmlDispatch)obj).getExpectedMsg(timeout);
    else if (obj instanceof KqmlRegistration)
      rval = ((KqmlRegistration)obj).getExpectedMsg(timeout);
    else if (obj instanceof KqmlAdmin)
      rval = ((KqmlAdmin)obj).getExpectedMsg(timeout);
    
    if (rval == null)
      {
	String message = null;
	long   tval    = -1;
	
	if (obj instanceof KqmlDispatch)
	  tval = ((KqmlDispatch)obj).getTimeout();
	else if (obj instanceof KqmlRegistration)
	  tval = ((KqmlRegistration)obj).getTimeout();
	else if (obj instanceof KqmlAdmin)
	  tval = ((KqmlAdmin)obj).getTimeout();
	
	message = "Timeout waiting for " + getKqmlMsgId();
	if (tval >= 0)
	  message += "\n     at " + formatter.format(new Date (tval));
	
	// Put this message on the list of messages which have timed out
	
	throw new KQMLMsgTimeout(message);
      }
    
    return (rval);
  }
  
  //
  // non-blocking sendAndAwaitReply that uses the default timeout value
  // for this agent
  //
  /**
   * Sends the message instance to the "receiver" agent and <b>waits</b>
   * for a reply <b>Message</b> object. This is also called a "synchronous"
   * send.  Agent destination address lookup is performed by this method.
   * The agent's default timeout value is used for timeout.
   *
   * @exception kqml.KqmlIOException  if an I/O exception is generated sending
   *                                  the message.  Usually due to the remote
   *                                  end of the socket closing.
   * @exception kqml.InvalidKQMLMsgException    due to a null performative
   * @exception kqml.InvalidKQMLTokenException  due to invalid token in the
   *                                            input message
   * @exception kqml.KQMLMsgTimeout             message reception timeout
   */
  public Message sendAndAwaitReply()
       throws kqml.KqmlIOException, kqml.KQMLMsgTimeout,
	 kqml.InvalidKQMLTokenException,
	 kqml.InvalidKQMLMsgException
  {
    return (sendAndAwaitReply(agent.getTimeout()));
  }
  
  /**
   * Generates and returns the object according the coding in the
   * <code>:kril-wrapper</code> field of the message.
   *
   * @exception kqml.InvalidKrilWrapperException  generated if the incoming
   * message's <code>:kril-wrapper</code> is malformed.
   */
  public Object getMessageObject()
       throws kqml.InvalidKrilWrapperException
  {
    return (new KrilWrapper(this).getInputVal());
  }
  
  /**
   * Clears out a message's values data values, except for the
   * agent, so that it may be re-used.
   */
  public void disposeMessage ()
  {
    keywordValues = new Hashtable();
    performative  = "";
    connect       = null;
  }
  
  //
  // Blocking version of the above, for use by KqmlAgent constructor
  // when registering the agent with the ANS.
  //
  protected Message blockingSendAndAwaitReply()
       throws kqml.KqmlIOException, kqml.KQMLMsgTimeout,
	 kqml.InvalidKQMLTokenException, kqml.InvalidKQMLMsgException
  {
    Message rval  = null;
    Date    start = null;
    
    if (connect == null)
      connect = agent.lookupAgent(getReceiver());
    
    sendText(WITH_REPLY, null);
    Log.log(Log.WRITE+Log.DEBUG,
	    "blockingSendAndAwaitReply: following sendText");
    
    rval = new Message(connect);
    if (agent.timingEnabled())
      start = new Date();
    rval.readMessage();
    if (agent.timingEnabled())
      agent.setKmsgTime(getKqmlMsgId(), TimingRecord.READ, start,
			new Date());
    
    return (rval);
  }
  
  //
  // This function is used internally, following return from a
  // Performative Handler to transmit the performative's response
  //
  protected void sendResponse(Reply reply)
  {
    if (getReplyWith() == null)
      {
	Log.log(Log.WRITE,
		"sendResponse: replyWith is null, no response sent");
      }
    else
      {
	try
	  {
	    Log.log(Log.WRITE+Log.DEBUG, "sendResponse: call replyString");
	    String text = this.replyString(reply);
	    
	    Log.log(Log.WRITE+Log.DEBUG,
		    "sendResponse: call Connection.write");
	    Date start = null;
	    if (agent.timingEnabled())
	      start = new Date();
	    connect.write (text);
	    if (agent.timingEnabled())
	      agent.setKmsgTime(getKqmlMsgId(), TimingRecord.WRITE, start,
                                new Date());
	  }
	//
	// Should we try to re-establish the connection, if it has dropped?
	//
	catch (KqmlIOException e)
	  {
	    Log.log(Log.ERROR, "sendResponse: " + e.getMessage());
	  }
	catch (InvalidKQMLMsgException e)
	  {
	    Log.log(Log.ERROR, "sendResponse: " + e.getMessage());
	  }
	Log.log(Log.WRITE+Log.DEBUG, "sendResponse: return");
      }
  }
  
  //
  // TIMING is NOT recorded for user's "send" calls (NO REPLY IS EXPECTED)
  //
  private void sendText(boolean replyWanted, Object obj)
       throws KqmlIOException, InvalidKQMLMsgException
  {
    Date start = null;
    if (replyWanted && agent.timingEnabled())
      start = new Date();
    
    String text = this.asString(replyWanted);
    
    if (replyWanted)
      {
	//
	// Register this message as a "pending" message.
	// MUST done before the message is sent by "connect.write".
	//
	if (obj instanceof KqmlDispatch)
	  ((KqmlDispatch)obj).makePending(this);
	else if (obj instanceof KqmlRegistration)
	  ((KqmlRegistration)obj).makePending(this);
	else if (obj instanceof KqmlAdmin)
	  ((KqmlAdmin)obj).makePending(this);
      }

    // Print the outgoing KQML message
    Log.log(Log.WRITE, "Outgoing KQML Message:");
    Log.log(Log.WRITE, this);
    connect.write (text);
    
    if (replyWanted && agent.timingEnabled())
      agent.setKmsgTime(getKqmlMsgId(), TimingRecord.WRITE, start,
			new Date());
    
    Log.log(Log.WRITE, "sendText: after write to " +
	    connect.getRemoteAgent() + "\n     (perf=`" + performative +
	    "' length=" + text.length() + ")");
  }
  
  private synchronized String generateKmsgId()
  {
    return (agentId + generatedMsgNo++);
  }
  
  //
  // The following variables are used by readMessage during message parsing.
  //
  private Token        currentToken   = new Token();
  private char         inChar;
  private boolean      recordingValue = false;
  private StringBuffer valueBuffer;
  
  protected void readMessage()
       throws kqml.InvalidKQMLTokenException, kqml.InvalidKQMLMsgException
  {
    // The incoming message text is assumed to be of the form:
    // '(' performative_name [keyword_name keyword_value]* ')'
    // this method is to parse this text, extracting the various components
    // of the message, saving the values as appropriate.
    String keyName;
    String value;
    
    if (DEBUG)
      Log.log(Log.MEMORY, "rdMsg");
    
    recordingValue = false;
    
    // Every message must start with a '(' preceded by an optional
    // "start token".
    getNextToken();
    
    if (DEBUG)
      Log.log(Log.MEMORY, "paren");
    
    if (currentToken.getType() == Token.START_TOKEN)
      getNextToken();
    if (currentToken.getType() != Token.LEFT_PAREN)
      {
	throw new InvalidKQMLTokenException(
                "'(' expected at the beginning of the KQML message: " +
                "'" + currentToken.getValue () + "' " +
                "(" + currentToken.getType () + ")");
      }
    
    // The next token in every message is the name of the performative
    getNextToken();
    
    if (DEBUG)
      Log.log(Log.MEMORY, "perf");
    
    if (currentToken.getType() != Token.WORD)
      {
	if (currentToken.getType() == Token.LEFT_PAREN)
	  // Put the left paren back in the input buffer, because it
	  // may be the start of another message.
	  connect.restoreInBuffer ('(');
	
	connect.disposeOfInMessage();
	
	throw new InvalidKQMLTokenException(
                "performative name expected: " +
                "'" + currentToken.getValue() + "' " +
                "(" + currentToken.getType() + ")");
      }
    this.setPerformative(currentToken.getValue());
    
    // Following the performative name is an optional list of keyword/value
    // pairs.
    getNextToken ();
    
    if (DEBUG)
      Log.log(Log.MEMORY, "keywd");
    
    while (currentToken.getType () != Token.RIGHT_PAREN
	   && currentToken.getType () != Token.UNDEFINED)
      {
	if (currentToken.getType () != Token.WORD)
	  {
	    if (currentToken.getType () == Token.LEFT_PAREN)
	      // Put the left paren back in the input buffer, because
	      // it may be the start of another message.
	      connect.restoreInBuffer('(');
	    connect.disposeOfInMessage();
	    throw new InvalidKQMLTokenException( "invalid keyword name: " +
                      "'" + currentToken.getValue() + "' " +
                      "(" + currentToken.getType() + ")");
	  }
	
	keyName = currentToken.getValue();
	
	recordingValue = true;
	valueBuffer    = new StringBuffer();
	
	getNextToken();
	if (DEBUG)
	  Log.log(Log.MEMORY, "value token");
	readExpression();
	
	if (DEBUG)
	  Log.log(Log.MEMORY, "readEx");
	if (DEBUG)
	  Log.log(Log.DEBUG, "valueBuffer length "+ valueBuffer.length());
	
	recordingValue = false;
	
	setKeywordValue(keyName.toLowerCase(), valueBuffer.toString());
	valueBuffer = null;
	if (DEBUG)
	  Log.log(Log.MEMORY, "addArg");
	
	getNextToken();
	
	if (DEBUG)
	  Log.log(Log.MEMORY, "next keywd");
      }
    
    // Every message must end with a ')'
    if (currentToken.getType() != Token.RIGHT_PAREN)
      {
	if (currentToken.getType() == Token.LEFT_PAREN)
	  // Put the left paren back in the input buffer, because
	  // it may be the start of another message.
	  connect.restoreInBuffer('(');
	
	connect.disposeOfInMessage();
	
	throw new InvalidKQMLTokenException(
                "')' expected at the end of the KQML message: " +
                "'" + currentToken.getValue() + "' " +
                "(" + currentToken.getType() + ")");
      }
    connect.finishBuffer();
  }
  
  private void readExpression()
       throws kqml.InvalidKQMLTokenException, kqml.InvalidKQMLMsgException
  {
    // NOTE: this method assumes that currentToken already contains the
    // first token in the expression
    switch (currentToken.getType())
      {
      case Token.WORD:
      case Token.STRING:
      case Token.BYTESTRING:
	break;
      case Token.LEFT_PAREN:
	readSExpression();
	break;
      case Token.TICK:
	getNextToken();
	readExpression();
	break;
      case Token.BACKTICK:
	getNextToken();
	if (currentToken.getType() == Token.COMMA)
	  readCommaExpression();
	else
	  readExpression();
	break;
      default:
	if (currentToken.getType() == Token.LEFT_PAREN)
	  // Put the left paren back in the input buffer, because
	  // it may be the start of another message.
	  connect.restoreInBuffer('(');
	
	connect.disposeOfInMessage();
	
	throw new InvalidKQMLTokenException(
                            "invalid keyword expression token: " +
                            "'" + currentToken.getValue () + "' " +
                            "(" + currentToken.getType () + ")");
      }
  }
  
  private void readExpressionOrCommaExpression()
       throws kqml.InvalidKQMLTokenException, kqml.InvalidKQMLMsgException
  {
    getNextToken();
    
    if (currentToken.getType() == Token.COMMA)
      readCommaExpression();
    else
      readExpression();
  }
  
  private void readSExpression()
       throws kqml.InvalidKQMLTokenException, kqml.InvalidKQMLMsgException
  {
    getNextToken();
    
    while (true)
      {
	switch (currentToken.getType())
	  {
	  case Token.RIGHT_PAREN:
	    return;
	  case Token.COMMA:
	    readCommaExpression();
	    break;
	  default:
	    readExpression();
	    break;
	  }
	getNextToken();
      }
  }
  
  private void readCommaExpression()
       throws kqml.InvalidKQMLTokenException, kqml.InvalidKQMLMsgException
  {
    getNextToken();
    
    switch (currentToken.getType())
      {
      case Token.RIGHT_PAREN:
	readSExpression();
	break;
      case Token.WORD:
	break;
      default:
	if (currentToken.getType() == Token.LEFT_PAREN)
	  // Put the left paren back in the input buffer, because it
	  // may be the start of another message.
	  connect.restoreInBuffer('(');
	
	connect.disposeOfInMessage();
	
	throw new InvalidKQMLTokenException (
                                "invalid keyword expression token: " +
                                "'" + currentToken.getValue() + "' " +
                                "(" + currentToken.getType() + ")");
      }
  }
  
  private void getNextToken ()
       throws kqml.InvalidKQMLTokenException, kqml.InvalidKQMLMsgException
  {
    int          tempIndex;
    int          tokenHdrLength = 0;
    int          tokenLength;
    StringBuffer tokenString    = new StringBuffer();
    
    inChar = connect.nextInChar();
    
    //
    // Skip over leading whitespace
    //
    boolean whitespaceFound = false;
    while (Character.isWhitespace(inChar))
      {
	whitespaceFound = true;
	inChar = connect.nextInChar();
      }
    if (whitespaceFound && recordingValue && valueBuffer.length() > 0)
      {
	valueBuffer.append(' ');
      }
    
    switch (inChar)
      {
      case '(':
	currentToken.setToken(Token.LEFT_PAREN, "(");
	break;
	
      case ')':
	currentToken.setToken(Token.RIGHT_PAREN, ")");
	break;
	
      case '\'':
	currentToken.setToken(Token.TICK, "'");
	break;
	
      case '`':
	currentToken.setToken(Token.BACKTICK, "`");
	break;
	
      case ',':
	currentToken.setToken(Token.COMMA, ",");
	break;
	
      case '#':
	//
	// The token is assumed to be of the form: #nn".....
	// where nn is the length of the text following the quote.
	//
	String lengthString = "";
	int lengthValue     = 0;
	
	inChar = connect.nextInChar();
	while (Character.isDigit(inChar))
	  {
	    lengthString += inChar;
	    inChar = connect.nextInChar();
	  }
	if (inChar != '"')
	  {
	    // If the double quote doesn't immediately follow a string
	    // of digits, the input is in error.
	    
	    // Put the "non-double quote" char read at the end of the
	    // previous loop back in the input buffer.
	    connect.restoreInBuffer (inChar);
	    throw new InvalidKQMLTokenException ("missing quote after '#'");
	  }
	
	try
	  {
	    lengthValue = Integer.parseInt(lengthString);
	  }
	catch (NumberFormatException e)
	  {
	    throw new InvalidKQMLTokenException("invalid length after '#'");
	  }
	
	tokenString.append(connect.nextInChars(lengthValue));
	
	if (recordingValue && valueBuffer.length() == 0)
	  currentToken.setToken(Token.BYTESTRING, tokenString.toString());
	else
	  currentToken.setToken (Token.BYTESTRING,
                        "#" + lengthString + "\"" + tokenString.toString());

	break;
	
      case '"':
	boolean isEscaped = false;  // used to test for the presence of '\\'
	
	// The token is assumed to be of the form "...." with only
	// "escaped" double quotes in the middle of the string
	inChar = connect.nextInChar();
	while (inChar != '"' || isEscaped)
	  {
	    tokenString.append(inChar);
	    isEscaped = (inChar == '\\');
	    inChar    = connect.nextInChar();
	  }
	
	if (recordingValue && valueBuffer.length() == 0)
	  currentToken.setToken(Token.STRING, tokenString.toString());
	else
	  currentToken.setToken(Token.STRING,
				"\"" + tokenString.toString() + "\"");
	break;
	
      default:
	if (Token.isWordChar(inChar))
	  {
	    while (Token.isWordChar(inChar))
	      {
		tokenString.append(inChar);
		inChar = connect.nextInChar();
	      }
	    
	    if (inChar == '#' & tokenString.toString().equals ("!@"))
	      {
		// The first 3 characters of the 5 character "start token"
		// (!@#$%) has been read.  See if the last 2 characters
		// also match)
		inChar = connect.nextInChar();
		if (inChar == '$')
		  {
		    inChar = connect.nextInChar();
		    if (inChar == '%')
		      {
			currentToken.setToken(Token.START_TOKEN,
					      START_TOKEN_STRING);
		      }
		    else
		      {
			connect.restoreInBuffer(inChar);
			connect.restoreInBuffer('$');
			connect.restoreInBuffer('#');
			currentToken.setToken(Token.WORD,
					      tokenString.toString());
		      }
		  }
		else
		  {
		    connect.restoreInBuffer(inChar);
		    connect.restoreInBuffer('#');
		    currentToken.setToken(Token.WORD,
					  tokenString.toString());
		  }
	      }
	    else
	      {
		// Put the extra non-word char read in the previous loop
		// back in the input buffer.
		connect.restoreInBuffer(inChar);
		
		currentToken.setToken(Token.WORD, tokenString.toString());
	      }
	  }
	else
	  {
	    throw new InvalidKQMLTokenException(
                      "unknown character in message: '" + inChar + "'");
	  }
	break;
      }
    
    tokenString = null;
    if (recordingValue)
      valueBuffer.append(currentToken.getValue());
    
    // System.out.println ("NEXT TOKEN: " + currentToken.getType () +
    //                     " -- " +  currentToken.getValue () + "\n");
  }
  
  private static final int MAX_SHORT_STRING_LENGTH = 100;
  
  protected static String quoteString(String text)
  {
    // change the specified text into a "KQML string" using the appropriate
    // form of KQML quoting (either "...." or #nn"....).

    // If the text is a Lisp s-expression, then pass it through as is
    if (isSexpr(text))
	return (text);
    else if (text.length() <= MAX_SHORT_STRING_LENGTH &&
	     text.indexOf('"') == -1)
      //
      // Simply put double quotes around "short" strings that have no
      // embedded double quotes.
      //
      return ("\"" + text + "\"");
    else
      // Otherwise, use the #nn" form of quoted string
      return ("#" + text.length() + "\"" + text);
  }
  
  private String asString(boolean replyWanted)
       throws kqml.InvalidKQMLMsgException
  {
    //
    // The message object is transformed into an (ascii) String
    // that external KQML agents can process.
    //
    boolean      kqmlMsgIdSeen   = false;
    boolean      replyWithSeen   = false;
    StringBuffer textBuffer      = new StringBuffer("(");
    String name;
    String value;
    
    if (performative != null && performative.length() > 0)
      textBuffer.append(performative);
    else
      throw new InvalidKQMLMsgException (PERF_ERROR);
    
    for (Enumeration keys = keywordValues.keys(); keys.hasMoreElements();)
      {
	name  = (String)keys.nextElement();
	value = getKeywordValue(name);
	
	if (value != null && value.length() > 0)
	  textBuffer.append(" " + name + " " + quoteString(value));
	
	if (name.equalsIgnoreCase(REPLY_WITH))
	  replyWithSeen = true;
	else if (name.equalsIgnoreCase(KQML_MSG_ID))
	  kqmlMsgIdSeen = true;
      }
    
    // Generate :reply-with if none present and a reply was wanted 
    if (replyWithSeen == false && replyWanted == true)
      textBuffer.append(" " + REPLY_WITH + " T");
    
    //
    // Generate :kqml-msg-id if none present, AND must add it
    // to the keyword value pairs in this message or it can't 
    // used for reply matching.
    //
    if (kqmlMsgIdSeen == false)
      {
	String kqmlMsgId = generateKmsgId();
	if (kqmlMsgId != null && kqmlMsgId.length() > 0)
	  textBuffer.append(" " + KQML_MSG_ID + " " +
			    quoteString(kqmlMsgId));
	setKeywordValue(KQML_MSG_ID, kqmlMsgId);
      }
    
    textBuffer.append(")");
    
    return new String(textBuffer);
  }
  
  protected String replyString(Reply reply)
       throws kqml.InvalidKQMLMsgException
  {
    //
    // The text is generated for the specified reply using this
    // message instance's SENDER, RECEIVER, REPLY_WITH, and KQML_MSG_ID
    // fields in the reply.
    //
    String       replyPerformative  = reply.getPerformative();
    StringBuffer textBuffer         = new StringBuffer("(");
    String       value;
    
    if (replyPerformative != null && replyPerformative.length() > 0)
      textBuffer.append(replyPerformative);
    else
      throw new InvalidKQMLMsgException(PERF_ERROR);
    
    //
    // These four fields values are fetched from the incoming message's
    // instance;  the sender is made the receiver, the receiver the
    // the sender, the reply-with is made an in-reply-to and the
    // kqml-msg-id is returned as is
    //
    value = getReceiver();
    if (value != null && value.length() > 0)
      textBuffer.append(" " + SENDER + " " + quoteString(value));
    
    value = getSender();
    if (value != null && value.length() > 0)
      textBuffer.append(" " + RECEIVER + " " + quoteString(value));
    
    value = getReplyWith();
    if (value != null && value.length() > 0)
      textBuffer.append(" " + IN_REPLY_TO + " " + quoteString(value));
    
    value = getKqmlMsgId();
    if (value != null && value.length() > 0)
      textBuffer.append(" " + KQML_MSG_ID + " " + quoteString(value));
    
    reply.generateMessageText(textBuffer);
    
    textBuffer.append(")");
    
    return (new String(textBuffer));
  }

  protected static boolean isSexpr(String str)
  {
    String beginParen = "(";
    String endParen   = ")";
    int    numBeginParens = 0;
    int    numEndParens   = 0;

    // If the string begins with a '(' and ends with a ')' and ...
    if (str.startsWith(beginParen)  &&  str.endsWith(endParen))
      {
	char c;
	
	// ... if the beginning and ending parens are balanced ...
	for (int i = 0 ; i < str.length() ; i++)
	  {
	    c = str.charAt(i);
	    if (c == '(')
	      numBeginParens++;
	    if (c == ')')
	      numEndParens++;
	  }
	if (numBeginParens == numEndParens)
	  {
	    // ... then it's an s-expression
	    return true;
	  }
      }
    return false;
  }
}

/**
 * A class to construct a token used in parsing incoming KQML messages.
 *
 * @author Michael Horton
 * @version $Revision: 1.7 $
 */
class Token
{
  protected static final int UNDEFINED    = 0;
  protected static final int LEFT_PAREN   = 1;
  protected static final int RIGHT_PAREN  = 2;
  protected static final int WORD         = 3;
  protected static final int STRING       = 4;
  protected static final int BYTESTRING   = 5;
  protected static final int COMMA        = 6;
  protected static final int TICK         = 7;
  protected static final int BACKTICK     = 8;
  protected static final int START_TOKEN  = 9;
  
  private int    tokenType;
  private String tokenValue;
  
  // Constructors
  Token()
  {
    tokenType  = UNDEFINED;
    tokenValue = new String();
  }
  
  // Methods
  protected void setToken (int tokenType, String tokenValue)
  {
    this.tokenType  = tokenType;
    this.tokenValue = tokenValue;
  }
  protected int getType()
  {
    return this.tokenType;
  }
  
  protected String getValue()
  {
    return this.tokenValue;
  }

  protected static boolean isWordChar(char c)
  {
    //
    // Determine whether the specified character is one of KQML's
    // "word" characters.
    //
    switch (c)
      {
	//
	// Check for all the "special cases"
	//
      case '-':
      case '<':
      case '>':
      case '=':
      case '+':
      case '*':
      case '/':
      case '&':
      case '^':
      case '~':
      case '_':
      case '@':
      case '$':
      case '%':
      case ':':
      case '.':
      case '!':
      case '?':
	return true;
      default:
	// If not a "special case," is it a letter or digit?
	return (Character.isUpperCase (c) ||
		Character.isLowerCase (c) ||
		Character.isDigit (c));
      }
  }
}

