/* File: KqmlWatcher.java
 * Contains: A watcher for significant Kqml messages
 * Author: Jeff Dalton <J.Dalton@ed.ac.uk>
 * Created: January 1998
 * Updated: Tue Jan 27 21:05:48 1998 by Jeff Dalton
 * Copyright: (c) 1998, AIAI, University of Edinburgh
 */

package oplan.acp3;

import java.util.*;
import oplan.util.*;
import java.io.*;

import kqml.*;


/**
 * A KqmlWatcher is used to look for significant Kqml messages and
 * tell an implementation of ProcessViewer about them.
 *
 * @see ProcessViewer
 * @see MessageWatcher
 */

public class KqmlWatcher {

    protected ProcessViewer viewer;
    public    KqmlAgent kqmlAgent;
    protected MessageQueue kqmlEvents = new MessageQueue();

    /**
     * Our name as a Kqml agent
     */
    protected static String kqmlAgentName = "ACP3";

    /**
     * Create a KqmlWatcher connected to a ProcessViewer.
     */
    public KqmlWatcher(ProcessViewer viewer) {
	this.viewer = viewer;
    }

    /**
     * Change the ProcessViewer that the KqmlWatcher tells of status changes.
     */
    public void setViewer(ProcessViewer viewer) {
	this.viewer = viewer;
    }

    /**
     * The init() method creates a KqmlAgent and then registers message
     * handlers for the Kqml performatives we care about.
     */
    public void init() {

	Debug.assert(viewer instanceof ProcessViewer, "no viewer");

	Debug.noteln("Creating Kqml agent");
	// Create the Kqml agent.
	try {
	    kqmlAgent = new KqmlAgent(kqmlAgentName);
	}
        catch (KqmlAgentRegistrationFailure e) { fatalException(e); }
        catch (KqmlAgentInitFailure e)         { fatalException(e); }

	// Register handler(s).
	kqmlAgent.registerHandler("evaluate", new MessageWatcher(this));
	kqmlAgent.registerHandler("reply", new MessageWatcher(this));
	kqmlAgent.registerHandler("error", new MessageWatcher(this));
    }

    /**
     * Analyzes messages enqueued by pleaseAnalyze.
     */
    public void watcherMainLoop() {
	Debug.assert(viewer instanceof ProcessViewer, "no viewer");
	Debug.assert(kqmlAgent instanceof KqmlAgent, "no Kqml agent");

	Debug.noteln("Starting watcherMainLoop");
	for (;;) {
	    Message msg = (Message)kqmlEvents.nextMessage();
	    Thread.yield();
	    analyzeMessage(msg);
	}
    }

    /**
     * Enqueues a Kqml Message for analysis in watcherMainLoop.
     */
    public void pleaseAnalyze(Message msg) {
	kqmlEvents.send(msg);	// here send means enqueue
    }

    /**
     * Performs the actual message analysis.
     */
    protected void analyzeMessage(Message msg) {
	Debug.noteln("Looking at", msg);

	String performative = msg.getPerformative().toLowerCase();
	String contents = msg.getContents().toLowerCase();

	// For now, the only Kqml performative we expect is "evaluate"
	if (!performative.equals("evaluate")) {
	    Util.warn("Unexpected Kqml performative: " + performative);
	    return;
	}

	// We're looking for :content of the form
	//   (:init-problem :task "as-a" :plan ?plan-advice)
	// where ?plan-advice is "plan-advice2" or "plan-advice3" (in quotes).
	// When we see one of those messages, we tell the viewer by calling
	// viewer.startAction.

	// We may also see "plan-advice1" or :content such as
	//   (:init-problem :task "as-a" :plan "plan-advice3-phase3"
	//    :action-network as-a-v1-15-post-ctem)
	// These we ignore.

	// See if the contents contains ":init-problem" but not "-phase".
	if (contents.startsWith(":init-problem", 1) 
	    && contents.indexOf("-phase") == -1) {

	    // Now see which COA, if any, this is about.
	    if (contents.indexOf("plan-advice2") != -1)
		viewer.startAction("develop COA", 2);
	    else if (contents.indexOf("plan-advice3") != -1)
		viewer.startAction("develop COA", 3);

	}

    }

    /**
     * shutdown should be called before the system exits so that it can
     * call kqmlAgent.shutdown().  The main program class should define a
     * shutdown method that calls our shutdown method, and ACP3.exit will
     * then cause the right things to happen.  That is, we should have
     * something like: ACP3.exit --> main program class shutdown -->
     * KqmlWatcher shutdown --> KqmlAgent shutdown.
     */
    public void shutdown() {
	Debug.noteln("KqmlWatcher shutdown");
	if (kqmlAgent != null) {
	    Debug.noteln("Calling kqml agent shutdown()");
	    kqmlAgent.shutdown();
	}
	else
	    Debug.noteln("No Kqml agent to shut down.");
    }

    /**
     * fatalException is called when an exception indicates that something
     * that must happen has failed, and there isn't anything in the system
     * that would enable it to fix the problem and try again.
     */
    static void fatalException(Exception e) {
        Debug.noteException(e);
	ACP3.exit(1);		// will arrange shutdown
    }

}


/*
 * Message handler classes
 */

class MessageWatcher implements Handler {

    protected KqmlWatcher watcher;

    /**
     * Create a MessageWatcher for a particular KqmlWatcher.
     */
    MessageWatcher(KqmlWatcher watcher) {
	this.watcher = watcher;
    }

    /**
     * Handles a Kqml message by enqueuing it for analysis by the
     * KqmlWatcher's watcherMainLoop, then returning a standard reply,
     * if a reply is required.
     *
     * @see MessageWatcher#makeReply
     */
    public Reply handler(Message msg) {
	Debug.noteln("Received Kqml:", msg);
	watcher.pleaseAnalyze(msg);
	// could instead do watcher.analyzeMessage(msg);
	return makeReply(msg);
    }

    /**
     * makeReply is responsible for making a "standard" reply to a message.
     * The assumption is that if we receive a message with :content of the
     * form (<plan-performative> ...), then (a) the :reply-with property
     * will be <plan-performative>, and (b) we can reply with Kqml
     * performative "reply" and content (<plan-performative> :ok).
     * Assumption (a) allows us to get the <plan-performative> from
     * the :reply-with property rather than from the :content.
     */
    static protected Reply makeReply(Message msg) {
	// Is a reply called for?
	// The :reply-with field is automatically send back as :in-reply-to.
	// It's not clear what we'll ever get :reply-with t or nil.
	String reply_with = msg.getReplyWith();
	// Debug.noteln("Making reply to", msg);
	if (reply_with == null)
	    return null;
	else if (reply_with.equalsIgnoreCase("nil"))
	    return null;
	else if (reply_with.equalsIgnoreCase("t"))
	    return Reply.reply(":ok");
	else
	    return Reply.reply("(" + reply_with + " :ok)");
    }

}
