/* File: MatchTest.java
 * Contains: Main program for testing Kqml message sending
 * Author: Jeff Dalton <J.Dalton@ed.ac.uk>
 * Created: January 1998
 * Updated: Tue Jan 27 20:07:40 1998 by Jeff Dalton
 * Copyright: (c) 1998, AIAI, University of Edinburgh
 */

package oplan.test;

import java.util.*;
import java.io.*;
import oplan.util.*;
import oplan.util.lisp.*;
import oplan.util.match.*;
import kqml.*;


/** A class to test Kqml message sending. */

class SendKqml {

    private static String ourname = "SendKqml";
    private static String serverName = "ACP3";
    private static KqmlAgent kqmlAgent = null;
    private static LispReader lispin = new LispReader(System.in);

    private static String[] performatives_to_handle = {
	"reply", "error", "sorry", "evaluate", "tell"
    };

    /* Symbols */

    static final Symbol K_CONTENT      = Symbol.intern(":content");
    static final Symbol S_EVALUATE     = Symbol.intern("evaluate");
    static final Symbol K_INIT_PROBLEM = Symbol.intern(":init-problem");

    static final Symbol S_PING         = Symbol.intern("ping");
    static final Symbol S_SENDLOG      = Symbol.intern("sendlog");
    static final Symbol S_HELP         = Symbol.intern("help");
    static final Symbol S_BYE          = Symbol.intern("bye");

    /* Help text */

    static final String[] helpText = {
	"",
	"Sending to " + serverName,
	"",
	"If you type a list, it should have the form",
	"  (<kqml-performative> ... :content <the-content> ...)",
	"",
	"Such a list will be converted to a Kqml Message and sent.",
	"Commands have the form: <name> <argument>*, and the following",
	"commands are available:",
	"",
	"   ping <agent-name>         ping the specified agent",
	"   sendlog <log-file-name>   send the contents of the file",
        "   help                      print this text",
	"   bye                       exit",
	""
    };

    /* Main */

    public static void main(String[] argv) {

	if (argv.length == 1) {
	    ourname = argv[0];
	}
	System.out.println("Calling ourself \"" + ourname + "\"");

	// Create the Kqml agent.
        try { kqmlAgent = new KqmlAgent(ourname); }
        catch (KqmlAgentRegistrationFailure e) { fatalException(e); }
        catch (KqmlAgentInitFailure e)         { fatalException(e); }

	// Register handlers for messages we think we might receive
	{
	    for (int i = 0; i < performatives_to_handle.length; i++) {
		String perf = performatives_to_handle[i];
		kqmlAgent.registerHandler
		    (perf, new MessageNoter(perf, kqmlAgent));
	    }
	}

	// If we don't lower our priority, replies aren't printed
	// while we're at the "Message:" prompt.  /\/ No, that helps
	// only some of the time, and the behaviour is different on
	// Solaris and FreeBSD.

	// Thread main_thread = Thread.currentThread();
	// main_thread.setPriority(main_thread.getPriority() - 1);

	// Loop asking the user for a message and sending it

	printLines(helpText);

    askloop:
	for (;;) {
	    try {
		Object input = askLisp("\nMessage:");
		if (input == S_BYE)
		    break askloop;
		else if (input instanceof Cons)
		    sendMessageFromList((Cons)input);
		else if (input == S_PING)
		    sendPing(lispin.readObject().toString());
		else if (input == S_SENDLOG)
		    sendLogFile(lispin.readObject().toString());
		else if (input == S_HELP)
		    printLines(helpText);
		else
		    System.out.println("I don't understand " + input);
	    }
	    catch (KqmlException e) {
		fatalException(e);
	    }
	    catch (RuntimeException e) {  // s.b. LispReadException
		Debug.noteException(e);
	    }
	}

	// And we're done.
	kqmlAgent.shutdown();	// needed?
	System.out.println("\nBye!\n");

    }

    private static void fatalException(Exception e) {
        System.out.println(e.getMessage());
	if (kqmlAgent != null)
	    kqmlAgent.shutdown();	// needed?
	System.exit(1);
    }

    static Object askLisp(String prompt) {
	System.out.print(prompt + " ");
	System.out.flush();
	// It's annoying that there's no lispin.available().
	try {
	    while (System.in.available() == 0) {
		try { Thread.sleep(1000); }
		catch (InterruptedException e) {}
	    }
	    return lispin.readObject();
	}
	catch (IOException e) {
	    Debug.noteException(e);
	    return Lisp.NIL;
	}

    }

    /* Message sending (from lists)
     *
     * A message might be:
     *
     * (EVALUATE :SENDER "pcm-1" :REPLY-WITH :INIT-PROBLEM :CONTENT
     *   (:INIT-PROBLEM :TASK "AS-A" :PLAN "plan-advice1"))
     *
     */

    private static void sendMessageFromList(List message) 
	 throws KqmlException
    {
	sendMessageFromList(message, true);
    }

    private static void quietlySendMessageFromList(List message) 
	 throws KqmlException
    {
	sendMessageFromList(message, false);
    }

    private static void sendMessageFromList(List message, boolean trace)
	 throws KqmlException
    {
        if (trace) Debug.noteln("Have:", Lisp.printToString(message));

	String performative = message.car().toString();
	List plist = message.cdr();
	String content = Lisp.printToString(plist.get(K_CONTENT));

	if (trace) Debug.noteln("Performative:", performative);
	if (trace) Debug.noteln("Content:", content);

	Message m = new Message(performative,
				content,
				serverName, 	// to
				kqmlAgent);	// from

	// Fill in message properties, skipping :content
	for (List p = plist; !p.isNull(); p = p.cdr().cdr()) {
	    Debug.assert(p.car() instanceof Symbol);
	    if (p.car() != K_CONTENT) {
		String prop = Lisp.printToString(p.car());
		String value = Lisp.printToString(p.cdr().car());
		if (trace) Debug.noteln("Property " + prop + " = " + value);
		m.setKeywordValue(prop, value);
	    }
	}

	// Send the message
	m.send();

    }


    /* Ping a specified agent */

    /* If you want to send a ping by hand, the list form of the message
     * looks like this:
     *
     *    (evaluate :reply-with :ping :content (:ping))
     *
     * Here we use sendMsgAndAwaitReply.
     */

    static void sendPing(String who) throws KqmlException {
	Message reply
	    = kqmlAgent.sendMsgAndAwaitReply("evaluate", "(:ping)", who);
	Debug.noteln("Reply:", reply);
    }


    /* Send the contents of a log file */

    // We look for (evaluate ... (:content <target>) ...)

    static Object target =
      Lisp.readFromString(
       "(:INIT-PROBLEM :TASK \"AS-A\" :PLAN ?plan-advice)");

    static void sendLogFile(String filename) throws KqmlException {
	LispReader lin;
	Object c;		// file contents

	System.out.println("Target = " + target);

	try {
	    lin = new LispFileReader(filename);
	    while ((c = safeRead(lin, Lisp.NIL)) != Lisp.EOF) {
		// Debug.noteln("Have:", c);
		// Util.askLine("ok?");
		analyseLogContent(c);
	    }
	    System.out.println("\nContents of " + filename + " sent.");
	}
	catch (FileNotFoundException e) {
	    Debug.noteException(e);
	}

    }

    static Object safeRead(LispReader lin, Object errValue) {
	try { return lin.readObject(); }
	catch (RuntimeException e) { 	// s.b. LispReadException
	    Debug.noteException(e);
	    return errValue;
	}
    }

    static void analyseLogContent(Object c) throws KqmlException {
	
	// Ignore atoms
	if (!(c instanceof Cons))
	    return;

	// See if the message looks significant.
	// First see if the Kqml performative is evaluate
	if (c instanceof Cons && ((Cons)c).car() == S_EVALUATE) {
	    List message = (List)c;
	    List plist = message.cdr();
	    Object content = plist.get(K_CONTENT);
	    MatchEnv env = SimpleMatcher.match(target, content);

	    // Note anything w/ content that starts (:init-problem ...)
	    if (content instanceof Cons
		&& ((Cons)content).car() == K_INIT_PROBLEM) {
		System.out.println("Maybe: " + c + "\n");
	    }

	    // If an actual match ...
	    if (env != null) {
		System.out.println("Found: " + c);
		System.out.println("Env = " + env + "\n");
		// Util.askLine("ok?");
	    }

	    // Send when Kqml performative = evaluate and content
	    // begins with :init-problem
	    if (content instanceof Cons
		&& ((Cons)content).car() == K_INIT_PROBLEM) {
		quietlySendMessageFromList(message);
	    }
	}

    }


    /* Print an array of lines, e.g. the help text. */

    static void printLines(String[] lines) {
	for (int i = 0; i < lines.length; i++) {
	    System.out.println(lines[i]);
	}
    }


}


/*
 * A simple handler class for messages we receive.
 */

class MessageNoter implements Handler {

    String performative;
    KqmlAgent kqmlAgent;

    MessageNoter(String performative, KqmlAgent kqmlAgent) {
	this.performative = performative;
	this.kqmlAgent = kqmlAgent;
    }

    public Reply handler(Message msg) {
	Debug.noteln("Received: " +
		     msg.getPerformative() +
		     ", content = " +
		     msg.getContents());
	Debug.noteln("Whole message: " + msg);
	if (msg.getContents().equals("die")) {
	    kqmlAgent.shutdown();
	    System.exit(0);
	}
	return makeReply(msg);
    }

    static protected Reply makeReply(Message msg) {
	// Is a reply called for?
	// The :reply-with field is automatically send back as :in-reply-to.
	// It's not clear what we'll ever get :reply-with t or nil.
	String reply_with = msg.getReplyWith();
	// Debug.noteln("Making reply to", msg);
	if (reply_with == null)
	    return null;
	else if (reply_with.equalsIgnoreCase("nil"))
	    return null;
	else if (reply_with.equalsIgnoreCase("t"))
	    return Reply.reply(":ok");
	else
	    return Reply.reply("(" + reply_with + " :ok)");
    }

}
